local Addon, _DUFO = ...
local strfind, strlen, strsub, strupper = string.find, string.len, string.sub, string.upper

local ElementClassSettings, ElementSpecSettings
local ClassFramesSaved = {}

local DUF_INIT_FRAMES = {}

local Wow_Project_Classic = WOW_PROJECT_CLASSIC or 2
local Wow_Project_Burning_Crusade_Classic = WOW_PROJECT_BURNING_CRUSADE_CLASSIC or 5
local Wow_Project_Wrath_Classic = WOW_PROJECT_WRATH_CLASSIC or 11
local RemoveOptions = { -- NS:IsValidVersion for Options
	DUF_Options = {
		{ frame="FocusFrame", [Wow_Project_Classic]=true, toplevel=true, },
		{ frame="BossFrame", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, toplevel=true,  },
		{ frame="DeathKnight", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, },
		{ frame="Mage", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, },
		{ frame="Monk", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, },
		{ frame="MonkStagger", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, },
		{ frame="Paladin", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, },
		{ frame="Warlock", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, },
		{ frame="Evoker", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, },
	},
}

local ReanchorOptions = {
	DUF_Options = {
		{ frame="DruidCombo", toframe="AltManaBarTxt", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, },
		{ frame="RogueCombo", toframe="DruidCombo", [Wow_Project_Classic]=true, [Wow_Project_Burning_Crusade_Classic]=true, [Wow_Project_Wrath_Classic]=true, },
	},
}

local function DUF_Options_GetAllSettings()
	if DUF_ELEMENT_INDEX and not strfind(DUF_ELEMENT_INDEX, "MONK") then
		ElementClassSettings = nil
		ElementSpecSettings = nil
	end
	if (DUF_ELEMENT_SUBINDEX) then
		return DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX][DUF_ELEMENT_SUBINDEX];
	elseif (DUF_ELEMENT_INDEX) then
		return DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX];
	else
		return DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX];
	end
end

local function DUF_Options_GetSetting(index)
	if (not DUF_FRAME_INDEX) then
		return DUF_Settings[DUF_INDEX][index];
	end
	if (DUF_ELEMENT_SUBINDEX) then
		return DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX][DUF_ELEMENT_SUBINDEX][index];
	elseif (DUF_ELEMENT_INDEX) then
		return DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX][index];
	else
		return DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][index];
	end
end

local function DUF_Run_Functions(self, initFunc)
	if (not DUF_INITIALIZED) then return; end
	if (not initFunc) then
		initFunc = self:GetParent().initFunc;
	end
	if (not initFunc) then
		initFunc = self.initFunc;
	end
	if (initFunc) then
		if (DUF_INIT_FRAMES) then
			for _,unit in pairs(DUF_INIT_FRAMES) do
				initFunc(self, unit, DUF_ELEMENT_SUBINDEX);
			end
		else
			initFunc(self);
		end
	end
end

local function DUF_Update_Setting(value, index, subindex, unusualsubindex)
	if (not index) then
		return;
	end
	if (not DUF_FRAME_INDEX) then
		if (subindex) then
			DUF_Settings[DUF_INDEX][index][subindex] = value;
		else
			DUF_Settings[DUF_INDEX][index] = value;
		end
	else
		if (DUF_ELEMENT_SUBINDEX) then
			DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX][DUF_ELEMENT_SUBINDEX][index] = value;
		elseif unusualsubindex then
			DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX][index][subindex][unusualsubindex] = value
		elseif subindex then
			DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX][index][subindex] = value
		elseif (DUF_ELEMENT_INDEX) then
			DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_ELEMENT_INDEX][index] = value;
		else
			DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][index] = value;
		end
	end
end

local function DUF_ColorPicker_ColorCancelled()
	local color = ColorPickerFrame.previousValues;
	DUF_Update_Setting(color, ColorPickerFrame.index, ColorPickerFrame.subindex);
	_G[ColorPickerFrame.colorBox]:SetBackdropColor(color.r, color.g, color.b);
	if (ColorPickerFrame.initfunc) then
		for _,unit in pairs(DUF_INIT_FRAMES) do
			ColorPickerFrame.initfunc(0, unit, DUF_ELEMENT_SUBINDEX);
		end
	end
end

local function DUF_ColorPicker_ColorChanged(self)
	local r, g, b = ColorPickerFrame:GetColorRGB();
	local color = { r=r, g=g, b=b };
	DUF_Update_Setting(color, ColorPickerFrame.index, ColorPickerFrame.subindex);
	_G[ColorPickerFrame.colorBox]:SetBackdropColor(color.r, color.g, color.b);
	if (ColorPickerFrame.initfunc) then
		for _,unit in pairs(DUF_INIT_FRAMES) do
			ColorPickerFrame.initfunc(self, unit, DUF_ELEMENT_SUBINDEX);
		end
	end
end

local function DUF_EditBox_UpdateFrameLoc(locIndex, value)
	local index = tonumber(DUF_FrameOptions_FrameChooser_Setting:GetText())
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][locIndex][index] = value
	if (strfind(DUF_FRAME_INDEX, "party")or strfind(DUF_FRAME_INDEX, "boss")) then
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX..index)
	else
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX)
	end
end

local function DUF_Init_BuffOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_BuffOptions_Size, settings.size);
	DL_Init_Slider(DUF_BuffOptions_HSpacing, settings.hspacing);
	DL_Init_Slider(DUF_BuffOptions_VSpacing, settings.vspacing);
	DL_Init_Slider(DUF_BuffOptions_Shown, settings.shown);
	DL_Init_CheckBox(DUF_BuffOptions_TwoRows, settings.tworows);
	DL_Init_CheckBox(DUF_BuffOptions_DurationFormat, settings.altformat);
	DL_Init_CheckBox(DUF_BuffOptions_Vertical, settings.vertical);
	DL_Init_CheckBox(DUF_BuffOptions_ReverseFill, settings.reverse);
	DL_Init_CheckBox(DUF_BuffOptions_ShowDuration, settings.showduration);
	DL_Init_CheckBox(DUF_BuffOptions_DurationFormat, settings.altformat);
	DL_Init_Slider(DUF_BuffOptions_DurationXOffset, settings.durationx);
	DL_Init_Slider(DUF_BuffOptions_DurationYOffset, settings.durationy);
	DL_Init_Slider(DUF_BuffOptions_DurationHeight, settings.durationheight);
	DL_Init_Slider(DUF_BuffOptions_DurationAlpha, settings.durationalpha);
	DL_Init_EditBox(DUF_BuffOptions_DurationFont, settings.durationfont);
	DL_Init_ColorPicker(DUF_BuffOptions_DurationColor, settings.durationcolor);
end

local function DUF_Init_TempEnchantOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_TempEnchantOptions_Size, settings.size);
	DL_Init_Slider(DUF_TempEnchantOptions_HSpacing, settings.hspacing);
	DL_Init_Slider(DUF_TempEnchantOptions_VSpacing, settings.vspacing);
	DL_Init_CheckBox(DUF_TempEnchantOptions_DurationFormat, settings.altformat);
	DL_Init_CheckBox(DUF_TempEnchantOptions_Vertical, settings.vertical);
	DL_Init_CheckBox(DUF_TempEnchantOptions_ShowDuration, settings.showduration);
	DL_Init_CheckBox(DUF_TempEnchantOptions_DurationFormat, settings.altformat);
	DL_Init_Slider(DUF_TempEnchantOptions_DurationXOffset, settings.durationx);
	DL_Init_Slider(DUF_TempEnchantOptions_DurationYOffset, settings.durationy);
	DL_Init_Slider(DUF_TempEnchantOptions_DurationHeight, settings.durationheight);
	DL_Init_Slider(DUF_TempEnchantOptions_DurationAlpha, settings.durationalpha);
	DL_Init_EditBox(DUF_TempEnchantOptions_DurationFont, settings.durationfont);
	DL_Init_ColorPicker(DUF_TempEnchantOptions_DurationColor, settings.durationcolor);
end

local function DUF_Init_ClassIconOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_ClassIconOptions_Size, settings.size);
	DL_Init_CheckBox(DUF_ClassIconOptions_UseRaceTexture, settings.userace);
end

local function UpdateTextureCount(frame, settings)
	local total, currentMax
	if DUF_ELEMENT_INDEX == "MONK"..SPEC_MONK_BREWMASTER then -- Stagger
		currentMax = settings.numtextures
	else
	 	total = DUF_CLASS_BAR_SPECS[DUF_ELEMENT_INDEX].total
		if settings.numtextures == 2 then
			currentMax = total
		else
			currentMax = settings.numtextures -- 0 or 1
		end
	end
	frame.CooledAlpha:Hide()
	frame.CooldownAlpha:Hide()
	frame.CooldownTexture:Hide()
	frame.CooldownColor:Hide()
	if DUF_ELEMENT_INDEX == "DEATHKNIGHT" then
		if currentMax > 0 then
			frame.CooledAlpha:Show()
			frame.CooldownAlpha:Show()
			frame.CooldownTexture:Show()
			frame.CooldownColor:Show()
			if not settings.cooldownalpha then
				settings.cooledalpha = 0.65
				settings.cooldownalpha = 1
				settings.cooldowntexture = ""
				settings.cooldowncolor = {r=1, g=1, b=1}
			end
			DL_Init_Slider(frame.CooledAlpha, settings.cooledalpha)
			DL_Init_Slider(frame.CooldownAlpha, settings.cooldownalpha)
			DL_Init_EditBox(frame.CooldownTexture, settings.cooldowntexture)
			DL_Init_ColorPicker(frame.CooldownColor, settings.cooldowncolor)
		end
	end

	for i=1, 8 do -- NEED TO CHECK FOR MAX. ELEMENTS FOR ALL CLASSES
		local f = frame["Texture_"..i]
		if i > currentMax then
			settings.textures[i] = nil
			f:Hide()
		else
			if not settings.textures[i] then
				settings.textures[i] = ClassFramesSaved[DUF_ELEMENT_INDEX].textures[i]
			end
			DL_Set_Label(f, DUF_TEXT.Texture.." "..f.subindex)
			DL_Init_EditBox(f, settings.textures[i].texture)
			f:Show()
		end
	end
	frame.currentMax = currentMax
end

local function SetTextureCount(self)
	local settings = DUF_DropMenu.settings
	UpdateTextureCount(DUF_DropMenu.frame, settings)
end

function DUF_ClassTextureChanged(self, ...)
	ClassFramesSaved[DUF_ELEMENT_INDEX][self.index][self.subindex].texture = self:GetText()
end

local function DUF_Init_ClassBarOptions(frame, settings)
	DL_Init_MenuControl(frame.Direction, settings.direction)
	DL_Init_CheckBox(frame.CenterAlign, settings.centeralign)
	if DUF_CLASS_CENTERS[settings.attachto] then
		frame.CenterAlign:Show()
	else
		frame.CenterAlign:Hide()
	end

	frame.Width.synclock = frame.SizeLock
	frame.Width.syncframe = frame.Height
	frame.Height.synclock = frame.SizeLock
	frame.Height.syncframe = frame.Width
	DL_Init_CheckBox(frame.SizeLock, settings.sizelock)
	DL_Init_Slider(frame.Width, settings.width)
	DL_Init_Slider(frame.Height, settings.height)
	DL_Init_Slider(frame.OOCAlpha, settings.oocalpha)
	DL_Init_Slider(frame.Spacing, settings.spacing)
	DL_Init_ColorPicker(frame.Color, settings.color)
	if not frame.NumTextures.initFunc then
		frame.NumTextures.initFunc = SetTextureCount
	end
	DUF_DropMenu.frame = frame
	DUF_DropMenu.settings = settings
	DL_Init_MenuControl(frame.NumTextures, settings.numtextures)
	UpdateTextureCount(frame, settings)
	DUF_Options_Initialize_ClassBar(frame)
	if frame.CDTHide then
		if WOW_PROJECT_ID == WOW_PROJECT_WRATH_CLASSIC then
			DL_Init_CheckBox(frame.WrathDefaultRunes, settings.wrathdefaultrunes)
			frame.WrathDefaultRunes:Show()
		end
		DL_Init_CheckBox(frame.CDTHide, settings.cooldownfonthide)
		DL_Init_EditBox(frame.CDTFont, settings.cooldownfont)
		DL_Init_Slider(frame.CDTHeight, settings.cooldownfontsize)
		if frame.CDSpecColor then
			DL_Init_CheckBox(frame.CDSpecColor, settings.cooldownspeccolors)
		end
	end
end

local function DUF_Init_Options_Rogue()
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Rogue_Settings.OnShowFunc then
		DUF_ElementOptions_Rogue_Settings.OnShowFunc = DUF_Init_Options_Rogue
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Rogue_Settings, settings)	
end

local function DUF_Init_Options_Paladin()
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Paladin_Settings.OnShowFunc then
		DUF_ElementOptions_Paladin_Settings.OnShowFunc = DUF_Init_Options_Paladin
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Paladin_Settings, settings)	
end

local function DUF_Init_Options_Monk()
	ElementClassSettings = "MONK"
	ElementSpecSettings = SPEC_MONK_WINDWALKER
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Monk_Settings.OnShowFunc then
		DUF_ElementOptions_Monk_Settings.OnShowFunc = DUF_Init_Options_Monk
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Monk_Settings, settings)	
end

local function DUF_Init_Options_Monk_Stagger()
	ElementClassSettings = "MONK"
	ElementSpecSettings = SPEC_MONK_BREWMASTER
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Monk_Stagger_Settings.OnShowFunc then
		DUF_ElementOptions_Monk_Stagger_Settings.OnShowFunc = DUF_Init_Options_Monk_Stagger
	end
	DUF_ElementOptions_Monk_Stagger_Settings.NumTextures:Hide() -- for bars
	DUF_ElementOptions_Monk_Stagger_Settings.SizeLock:Hide()
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Monk_Stagger_Settings, settings)	
end

local function DUF_Init_Options_Mage()
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Mage_Settings.OnShowFunc then
		DUF_ElementOptions_Mage_Settings.OnShowFunc = DUF_Init_Options_Mage
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Mage_Settings, settings)	
end

local function DUF_Init_Options_DeathKnight()
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_DeathKnight_Settings.OnShowFunc then
		DUF_ElementOptions_DeathKnight_Settings.OnShowFunc = DUF_Init_Options_DeathKnight
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_DeathKnight_Settings, settings)
end

local function DUF_Init_Options_Evoker()
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Evoker_Settings.OnShowFunc then
		DUF_ElementOptions_Evoker_Settings.OnShowFunc = DUF_Init_Options_Evoker
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Evoker_Settings, settings)
end

local function DUF_Init_Options_Druid()
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Druid_Settings.OnShowFunc then
		DUF_ElementOptions_Druid_Settings.OnShowFunc = DUF_Init_Options_Druid
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Druid_Settings, settings)	
end

local function DUF_Init_Options_Warlock()
	local settings = DUF_Options_GetAllSettings()
	if not DUF_ElementOptions_Warlock_Settings.OnShowFunc then
		DUF_ElementOptions_Warlock_Settings.OnShowFunc = DUF_Init_Options_Warlock
	end
	DUF_Init_ClassBarOptions(DUF_ElementOptions_Warlock_Settings, settings)	
end

local function DUF_Init_TotemFrame()
	local frame = DUF_ElementOptions_TotemFrame_Settings
	local settings = DUF_Options_GetAllSettings()
	DL_Init_Slider(frame.Scale, settings.scale)
	DL_Init_MenuControl(frame.Direction, settings.direction)
	DL_Init_Slider(frame.Spacing, settings.spacing)
end
local function DUF_Init_TargetMarkersOptions()
	local settings = DUF_Options_GetAllSettings()
	DL_Init_Slider(DUF_TargetMarkersOptions_Size, settings.markersize)
	DL_Init_Slider(DUF_TargetMarkersOptions_Width, settings.width)
	DL_Init_Slider(DUF_TargetMarkersOptions_Height, settings.height)
	DL_Init_CheckBox(DUF_TargetMarkersOptions_HideWhenSolo,settings.hidewhensolo)
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos1, settings.xos[1])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos1, settings.yos[1])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos2, settings.xos[2])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos2, settings.yos[2])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos3, settings.xos[3])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos3, settings.yos[3])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos4, settings.xos[4])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos4, settings.yos[4])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos5, settings.xos[5])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos5, settings.yos[5])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos6, settings.xos[6])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos6, settings.yos[6])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos7, settings.xos[7])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos7, settings.yos[7])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Xpos8, settings.xos[8])
	DL_Init_EditBox(DUF_TargetMarkersOptions_Ypos8, settings.yos[8])
end

local function DUF_Init_DebuffOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_DebuffOptions_Size, settings.size);
	DL_Init_Slider(DUF_DebuffOptions_HSpacing, settings.hspacing);
	DL_Init_Slider(DUF_DebuffOptions_VSpacing, settings.vspacing);
	DL_Init_Slider(DUF_DebuffOptions_Shown, settings.shown);
	DL_Init_CheckBox(DUF_DebuffOptions_TwoRows, settings.tworows);
	DL_Init_CheckBox(DUF_DebuffOptions_Vertical, settings.vertical);
	DL_Init_CheckBox(DUF_DebuffOptions_Flash, settings.flash);
	DL_Init_CheckBox(DUF_DebuffOptions_ReverseFill, settings.reverse);
	DL_Init_CheckBox(DUF_DebuffOptions_ShowDuration, settings.showduration);
	DL_Init_CheckBox(DUF_DebuffOptions_DurationFormat, settings.altformat);
	DL_Init_Slider(DUF_DebuffOptions_DurationXOffset, settings.durationx);
	DL_Init_Slider(DUF_DebuffOptions_DurationYOffset, settings.durationy);
	DL_Init_Slider(DUF_DebuffOptions_DurationHeight, settings.durationheight);
	DL_Init_Slider(DUF_DebuffOptions_DurationAlpha, settings.durationalpha);
	DL_Init_EditBox(DUF_DebuffOptions_DurationFont, settings.durationfont);
	DL_Init_ColorPicker(DUF_DebuffOptions_DurationColor, settings.durationcolor);
end

function DUF_Init_BaseOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_EditBox(DUF_ElementOptions_Base_XOffset, settings.xoffset);
	DL_Init_EditBox(DUF_ElementOptions_Base_YOffset, settings.yoffset);
	DL_Init_EditBox(DUF_ElementOptions_Base_FrameLevel, settings.framelevel);
	DL_Init_EditBox(DUF_ElementOptions_Base_BGTexture, settings.bgtexture);
	DL_Init_MenuControl(DUF_ElementOptions_Base_FrameStrata, settings.framestrata);
	DL_Init_EditBox(DUF_ElementOptions_Base_CustomTooltip, settings.customtooltip);
	DL_Init_CheckBox(DUF_ElementOptions_Base_DisableMouse, settings.disablemouse);
	DL_Init_CheckBox(DUF_ElementOptions_Base_UseCustomTooltip, settings.usecustomtooltip);
	DL_Init_CheckBox(DUF_ElementOptions_Base_DisableTooltip, settings.disabletooltip);
	DL_Init_CheckBox(DUF_ElementOptions_Base_Flash, settings.flashlowhealth);
	DL_Init_MenuControl(DUF_ElementOptions_Base_MainColorContext, settings.context);
	DL_Init_MenuControl(DUF_ElementOptions_Base_BackgroundColorContext, settings.bgcontext);
	DL_Init_MenuControl(DUF_ElementOptions_Base_BorderColorContext, settings.bordercontext);
	DL_Init_MenuControl(DUF_ElementOptions_Base_LeftClick, settings.leftclick);
	DL_Init_MenuControl(DUF_ElementOptions_Base_RightClick, settings.rightclick);
	DL_Init_MenuControl(DUF_ElementOptions_Base_MiddleClick, settings.middleclick);
	DL_Init_MenuControl(DUF_ElementOptions_Base_AttachFrame, settings.attach);
	DL_Init_MenuControl(DUF_ElementOptions_Base_AttachPoint, settings.attachpoint);
	DL_Init_MenuControl(DUF_ElementOptions_Base_AttachTo, settings.attachto);
	DL_Init_CheckBox(DUF_ElementOptions_Base_Hide, settings.hide);
	DL_Init_CheckBox(DUF_ElementOptions_Base_Mouseover, settings.mouseover);
	DL_Init_CheckBox(DUF_ElementOptions_Base_HideOnMouseover, settings.hidemouseover);
	DL_Init_CheckBox(DUF_ElementOptions_Base_UseMouseoverColor, settings.usemouseovercolor);
	DL_Init_ColorPicker(DUF_ElementOptions_Base_MouseoverColor, settings.mouseovercolor);
	DL_Init_MenuControl(DUF_ElementOptions_Base_MouseoverGroup, settings.mouseovergroup);
	DL_Init_Slider(DUF_ElementOptions_Base_Alpha, settings.alpha);
	DL_Init_CheckBox(DUF_ElementOptions_Base_HideBG, settings.hidebg);
	DL_Init_ColorPicker(DUF_ElementOptions_Base_BGColor, settings.bgcolor);
	DL_Init_Slider(DUF_ElementOptions_Base_BGAlpha, settings.bgalpha);
	DL_Init_Slider(DUF_ElementOptions_Base_BGPadding, settings.bgpadding);
	DL_Init_ColorPicker(DUF_ElementOptions_Base_BorderColor, settings.bordercolor);
	DL_Init_MenuControl(DUF_ElementOptions_Base_BorderStyle, settings.backgroundstyle);
	DL_Init_Slider(DUF_ElementOptions_Base_BorderAlpha, settings.borderalpha);
end

local function DUF_Init_EliteTextureOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_EliteTextureOptions_Size, settings.size);
	DL_Init_CheckBox(DUF_EliteTextureOptions_FacesLeft, settings.faceleft);
	DL_Init_EditBox(DUF_EliteTextureOptions_CustomEliteTexture, settings.elitetexture);
	DL_Init_EditBox(DUF_EliteTextureOptions_CustomRareTexture, settings.raretexture);
end

local function DUF_Init_QuestBossIconOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_QuestBossIconOptions_Size, settings.size);
	DL_Init_EditBox(DUF_QuestBossIconOptions_QuestBossIconTexture, settings.texture);
end

local function DUF_Init_FrameOptions()
	local settings = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX];
	DL_Init_MenuControl(DUF_FrameOptions_MainColorContext, settings.context);
	DL_Init_MenuControl(DUF_FrameOptions_FrameChooser, 1);
	DL_Init_MenuControl(DUF_FrameOptions_BackgroundColorContext, settings.bgcontext);
	DL_Init_MenuControl(DUF_FrameOptions_BorderColorContext, settings.bordercontext);
	DL_Init_MenuControl(DUF_FrameOptions_LeftClick, settings.leftclick);
	DL_Init_MenuControl(DUF_FrameOptions_RightClick, settings.rightclick);
	DL_Init_MenuControl(DUF_FrameOptions_MiddleClick, settings.middleclick);
	DL_Init_CheckBox(DUF_FrameOptions_Hide, settings.hide);
	DL_Init_CheckBox(DUF_FrameOptions_DisableMouse, settings.disablemouse);
	DL_Init_CheckBox(DUF_FrameOptions_Flash, settings.flash);
	DL_Init_CheckBox(DUF_FrameOptions_ShowDefault, settings.showDefault);
	DL_Init_ColorPicker(DUF_FrameOptions_MouseoverColor, settings.mouseovercolor);
	DL_Init_CheckBox(DUF_FrameOptions_UseMouseoverColor, settings.usemouseovercolor);
	DL_Init_MenuControl(DUF_FrameOptions_MouseoverGroup, settings.mouseovergroup);
	DL_Init_Slider(DUF_FrameOptions_Height, settings.height);
	DL_Init_Slider(DUF_FrameOptions_Width, settings.width);
	DL_Init_Slider(DUF_FrameOptions_Scale, settings.scale);
	DL_Init_Slider(DUF_FrameOptions_Alpha, settings.alpha);
	DL_Init_CheckBox(DUF_FrameOptions_HideBG, settings.hidebg);
	DL_Init_ColorPicker(DUF_FrameOptions_BGColor, settings.bgcolor);
	DL_Init_Slider(DUF_FrameOptions_BGAlpha, settings.bgalpha);
	DL_Init_Slider(DUF_FrameOptions_BGPadding, settings.bgpadding);
	DL_Init_ColorPicker(DUF_FrameOptions_BorderColor, settings.bordercolor);
	DL_Init_MenuControl(DUF_FrameOptions_BorderStyle, settings.backgroundstyle);
	DL_Init_Slider(DUF_FrameOptions_BorderAlpha, settings.borderalpha);
	DL_Init_CheckBox(DUF_FrameOptions_DisableTooltip, settings.disabletooltip);
	DL_Init_CheckBox(DUF_FrameOptions_UseCustomTooltip, settings.usecustomtooltip);
	DL_Init_EditBox(DUF_FrameOptions_CustomTooltip, settings.customtooltip);
	DL_Init_EditBox(DUF_FrameOptions_XOffset, settings.xos[1]);
	DL_Init_EditBox(DUF_FrameOptions_YOffset, settings.yos[1]);
	DL_Init_CheckBox(DUF_FrameOptions_Connect, settings.connect);
	DL_Init_MenuControl(DUF_FrameOptions_ConnectMethod, settings.connectmethod);
	DL_Init_Slider(DUF_FrameOptions_Spacing, settings.spacing);
	DL_Init_MenuControl(DUF_FrameOptions_AttachPoint, settings.attachpoint[1]);
	DL_Init_MenuControl(DUF_FrameOptions_AttachTo, settings.attachto[1]);
end

local function DUF_Init_HealthBarOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_HealthBarOptions_Height, settings.height);
	DL_Init_Slider(DUF_HealthBarOptions_Width, settings.width);
	DL_Init_ColorPicker(DUF_HealthBarOptions_Color, settings.color);
	DL_Init_ColorPicker(DUF_HealthBarOptions_SecondaryColor, settings.color2);
	DL_Init_CheckBox(DUF_HealthBarOptions_Fill, settings.fill);
	DL_Init_CheckBox(DUF_HealthBarOptions_Fade, settings.fade);
	DL_Init_CheckBox(DUF_HealthBarOptions_ResizeMax, settings.resizemax);
	DL_Init_CheckBox(DUF_HealthBarOptions_InCombat, settings.incombat);
	DL_Init_CheckBox(DUF_HealthBarOptions_HealthUpdating, settings.healthupdating);
	DL_Init_CheckBox(DUF_HealthBarOptions_ManaUpdating, settings.manaupdating);
	DL_Init_EditBox(DUF_HealthBarOptions_CustomTexture, settings.customtexture);
	DL_Init_EditBox(DUF_HealthBarOptions_CustomTexture2, settings.customtexture2);
	DL_Init_MenuControl(DUF_HealthBarOptions_Direction, settings.direction);
end

local function DUF_Init_IconOptions()
	local settings = DUF_Options_GetAllSettings();
	if (DUF_ELEMENT_INDEX == "HappinessIcon") then
		DUF_IconOptions_AnimateUnhappyIcon:Show()
	else
		DUF_IconOptions_AnimateUnhappyIcon:Hide()
	end
	if DUF_ELEMENT_INDEX == "RaceIcon" or DUF_ELEMENT_INDEX == "RankIcon" then
		DUF_IconOptions_TargetIcon:Show()
	else
		DUF_IconOptions_TargetIcon:Hide()
	end
	DL_Init_Slider(DUF_IconOptions_Size, settings.size);
	DL_Init_CheckBox(DUF_IconOptions_Circle, settings.circle);
	DL_Init_CheckBox(DUF_IconOptions_TargetIcon, settings.targetIcon);
	DL_Init_CheckBox(DUF_IconOptions_AnimateUnhappyIcon, settings.animatePetUnhappy);
end

local function DUF_Init_ManaBarOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_ManaBarOptions_Height, settings.height);
	DL_Init_Slider(DUF_ManaBarOptions_Width, settings.width);
	DL_Init_ColorPicker(DUF_ManaBarOptions_ManaColor, settings.manacolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_RageColor, settings.ragecolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_FocusColor, settings.focuscolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_EnergyColor, settings.energycolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_SecondaryColor, settings.color2);
	DL_Init_ColorPicker(DUF_ManaBarOptions_RunicPowerColor, settings.runicpowercolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_InsanityColor, settings.insanitycolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_LunarPowerColor, settings.lunarpowercolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_MaelstromPowerColor, settings.maelstrompower);
	DL_Init_ColorPicker(DUF_ManaBarOptions_FuryColor, settings.furycolor);
	DL_Init_ColorPicker(DUF_ManaBarOptions_HolyColor, settings.holycolor);
	DL_Init_CheckBox(DUF_ManaBarOptions_Fill, settings.fill);
	DL_Init_CheckBox(DUF_ManaBarOptions_Fade, settings.fade);
	DL_Init_CheckBox(DUF_ManaBarOptions_InCombat, settings.incombat);
	DL_Init_CheckBox(DUF_ManaBarOptions_HealthUpdating, settings.healthupdating);
	DL_Init_CheckBox(DUF_ManaBarOptions_ManaUpdating, settings.manaupdating);
	DL_Init_CheckBox(DUF_ManaBarOptions_HideIfNoMana, settings.hideifnomana);
	DL_Init_CheckBox(DUF_ManaBarOptions_ResizeMax, settings.resizemax);
	DL_Init_EditBox(DUF_ManaBarOptions_CustomTexture, settings.customtexture);
	DL_Init_EditBox(DUF_ManaBarOptions_CustomTexture2, settings.customtexture2);
	DL_Init_MenuControl(DUF_ManaBarOptions_Direction, settings.direction);
end

local function DUF_Init_AltManaBarOptions() 
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_AltManaBarOptions_Height, settings.height);
	DL_Init_Slider(DUF_AltManaBarOptions_Width, settings.width);
	DL_Init_ColorPicker(DUF_AltManaBarOptions_ManaColor, settings.manacolor);
	DL_Init_ColorPicker(DUF_AltManaBarOptions_SecondaryColor, settings.color2);
	DL_Init_CheckBox(DUF_AltManaBarOptions_Fill, settings.fill);
	DL_Init_CheckBox(DUF_AltManaBarOptions_Fade, settings.fade);
	DL_Init_CheckBox(DUF_AltManaBarOptions_InCombat, settings.incombat);
	DL_Init_CheckBox(DUF_AltManaBarOptions_HealthUpdating, settings.healthupdating);
	DL_Init_CheckBox(DUF_AltManaBarOptions_ManaUpdating, settings.manaupdating);
	DL_Init_CheckBox(DUF_AltManaBarOptions_HideIfNoMana, settings.hideifnomana);
	DL_Init_CheckBox(DUF_AltManaBarOptions_ResizeMax, settings.resizemax);
	DL_Init_EditBox(DUF_AltManaBarOptions_CustomTexture, settings.customtexture);
	DL_Init_EditBox(DUF_AltManaBarOptions_CustomTexture2, settings.customtexture2);
	DL_Init_MenuControl(DUF_AltManaBarOptions_Direction, settings.direction);
end

local function DUF_Init_PortraitOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_PortraitOptions_Size, settings.size);
	DL_Init_CheckBox(DUF_PortraitOptions_UseClassTexture, settings.useclass);
	DL_Init_CheckBox(DUF_PortraitOptions_Flip, settings.flip);
	DL_Init_CheckBox(DUF_PortraitOptions_ShowBorder, settings.showborder);
	DL_Init_CheckBox(DUF_PortraitOptions_Use3DPortrait, settings.use3D);
	DL_Init_CheckBox(DUF_PortraitOptions_Use3DFace, settings.use3Dface);
	DL_Init_ColorPicker(DUF_PortraitOptions_BorderColor, settings.bordcolor);
end

local function DUF_Init_TextBoxOptions()
	local settings = DUF_Options_GetAllSettings();
	if (not settings) then return; end
	DUF_TextBoxOptions.TextVariables.LastIndex = nil
	DUF_TextBoxOptions.TextVariables.InsertVariable:Hide()
	DUF_VarList_Update(DUF_TextBoxOptions.TextVariables)

	DL_Init_Slider(DUF_TextBoxOptions_Height, settings.height);
	DL_Init_Slider(DUF_TextBoxOptions_Width, settings.width);
	DL_Init_EditBox(DUF_TextBoxOptions_Text, settings.text);
	DL_Init_EditBox(DUF_TextBoxOptions_Label, settings.label);
	DL_Init_EditBox(DUF_TextBoxOptions_MaxCharacters, settings.maxcharacters);
	DL_Init_MenuControl(DUF_TextBoxOptions_FormatHealth, settings.formatH);
	DL_Init_CheckBox(DUF_TextBoxOptions_ShortHealthFormat, settings.shortHealth);
	DL_Init_MenuControl(DUF_TextBoxOptions_FormatMana, settings.formatM);
	DL_Init_CheckBox(DUF_TextBoxOptions_ShortManaFormat, settings.shortMana);
	DL_Init_MenuControl(DUF_TextBoxOptions_JustifyH, settings.justify);
	DL_Init_MenuControl(DUF_TextBoxOptions_JustifyV, settings.justifyV);
	DL_Init_MenuControl(DUF_TextBoxOptions_TextOutline, settings.outline);
	DL_Init_EditBox(DUF_TextBoxOptions_Font, settings.font);
	DL_Init_ColorPicker(DUF_TextBoxOptions_Color, settings.textcolor);
	DL_Init_Slider(DUF_TextBoxOptions_FontHeight, settings.textheight);
	DL_Init_CheckBox(DUF_TextBoxOptions_HideIfNoMana, settings.hideifnomana);
	DL_Init_CheckBox(DUF_TextBoxOptions_HideIfNoText, settings.hideifnotext);
	DL_Init_CheckBox(DUF_TextBoxOptions_InCombat, settings.incombat);
	DL_Init_CheckBox(DUF_TextBoxOptions_HealthUpdating, settings.healthupdating);
	DL_Init_CheckBox(DUF_TextBoxOptions_ManaUpdating, settings.manaupdating);
	DL_Init_CheckBox(DUF_TextBoxOptions_VerticalText, settings.verttext);
	DL_Init_EditBox(DUF_TextBoxOptions_HideIfTextIs, settings.hideiftextis)
end

local function DUF_Init_XPBarOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_XPBarOptions_Height, settings.height);
	DL_Init_Slider(DUF_XPBarOptions_Width, settings.width);
	DL_Init_ColorPicker(DUF_XPBarOptions_Color, settings.color);
	DL_Init_ColorPicker(DUF_XPBarOptions_SecondaryColor, settings.color2);
	DL_Init_EditBox(DUF_XPBarOptions_CustomTexture, settings.customtexture);
	DL_Init_EditBox(DUF_XPBarOptions_CustomTexture2, settings.customtexture2);
	DL_Init_MenuControl(DUF_XPBarOptions_Direction, settings.direction);
	DL_Init_CheckBox(DUF_XPBarOptions_Reputation, settings.trackRep);
	if (DUF_FRAME_INDEX == "player") then
		DUF_XPBarOptions_Reputation:Show()
	else
		DUF_XPBarOptions_Reputation:Hide()
	end
end

--[[
local function DUF_Init_CastBarOptions()
	local settings = DUF_Options_GetAllSettings();
	DL_Init_Slider(DUF_CastBarOptions_Height, settings.height);
	DL_Init_Slider(DUF_CastBarOptions_Width, settings.width);
	DL_Init_ColorPicker(DUF_CastBarOptions_ManaColor, settings.manacolor);
	DL_Init_ColorPicker(DUF_CastBarOptions_RageColor, settings.ragecolor);
	DL_Init_ColorPicker(DUF_CastBarOptions_FocusColor, settings.focuscolor);
	DL_Init_ColorPicker(DUF_CastBarOptions_EnergyColor, settings.energycolor);
	DL_Init_ColorPicker(DUF_CastBarOptions_SecondaryColor, settings.color2);
	DL_Init_CheckBox(DUF_CastBarOptions_Fill, settings.fill);
	DL_Init_CheckBox(DUF_CastBarOptions_Fade, settings.fade);
	DL_Init_CheckBox(DUF_CastBarOptions_InCombat, settings.incombat);
	DL_Init_CheckBox(DUF_CastBarOptions_HealthUpdating, settings.healthupdating);
	DL_Init_CheckBox(DUF_CastBarOptions_ManaUpdating, settings.manaupdating);
	DL_Init_CheckBox(DUF_CastBarOptions_HideIfNoMana, settings.hideifnomana);
	DL_Init_CheckBox(DUF_CastBarOptions_ResizeMax, settings.resizemax);
	DL_Init_EditBox(DUF_CastBarOptions_CustomTexture, settings.customtexture);
	DL_Init_EditBox(DUF_CastBarOptions_CustomTexture2, settings.customtexture2);
	DL_Init_MenuControl(DUF_CastBarOptions_Direction, settings.direction);
end
]]--

local function DUF_Options_ResetIndices()
	if (DUF_OLD_ELEM_SUBINDEX) then
		DUF_ELEMENT_SUBINDEX = DUF_OLD_ELEM_SUBINDEX;
		DUF_OLD_ELEM_SUBINDEX = nil;
	end
	if (DUF_OLD_ELEM_INDEX) then
		DUF_ELEMENT_INDEX = DUF_OLD_ELEM_INDEX;
		DUF_OLD_ELEM_INDEX = nil;
	end
	if (DUF_OLD_FRAME_INDEX) then
		DUF_FRAME_INDEX = DUF_OLD_FRAME_INDEX;
		DUF_OLD_FRAME_INDEX = nil;
	end
end

local function DUF_Options_SetOldIndices()
	if (not DUF_OLD_ELEM_INDEX) then
		DUF_OLD_ELEM_INDEX = DUF_ELEMENT_INDEX;
	end
	DUF_ELEMENT_INDEX = nil;
	if (not DUF_OLD_ELEM_SUBINDEX) then
		DUF_OLD_ELEM_SUBINDEX = DUF_ELEMENT_SUBINDEX;
	end
	DUF_ELEMENT_SUBINDEX = nil;
	if (not DUF_OLD_FRAME_INDEX) then
		DUF_OLD_FRAME_INDEX = DUF_FRAME_INDEX;
	end
	DUF_FRAME_INDEX = nil;
end

local function DUF_Show_SpecialOptions()
	if (DUF_SPECIAL_OPTIONS) then
		_G[DUF_SPECIAL_OPTIONS]:Hide();
	end
	if (DUF_ELEMENT_INDEX == "StatusBar") then
		if (DUF_ELEMENT_SUBINDEX == 1 or DUF_ELEMENT_SUBINDEX == 4 or DUF_ELEMENT_SUBINDEX == 6) then
			DUF_SPECIAL_OPTIONS = "DUF_HealthBarOptions";
			DUF_Init_HealthBarOptions();
		elseif (DUF_ELEMENT_SUBINDEX == 2 or DUF_ELEMENT_SUBINDEX == 5) then
			DUF_SPECIAL_OPTIONS = "DUF_ManaBarOptions";
			DUF_Init_ManaBarOptions();
		elseif (DUF_ELEMENT_SUBINDEX == 3 or DUF_ELEMENT_SUBINDEX == 8) then -- XP and Honor bar settings
			DUF_SPECIAL_OPTIONS = "DUF_XPBarOptions";
			DUF_Init_XPBarOptions();
--		elseif(DUF_ELEMENT_SUBINDEX == 6)  then  -- Fixed 5.3
--			DUF_SPECIAL_OPTIONS = "DUF_CastBarOptions";
--			DUF_Init_CastBarOptions();
		elseif (DUF_ELEMENT_SUBINDEX == 7) then -- Fixed 5.3
			DUF_SPECIAL_OPTIONS = "DUF_AltManaBarOptions";
			DUF_Init_AltManaBarOptions();
		end
	elseif (DUF_ELEMENT_INDEX == "StatusIcon" or DUF_ELEMENT_INDEX == "RankIcon" or DUF_ELEMENT_INDEX == "RaceIcon" or DUF_ELEMENT_INDEX == "PVPIcon" or DUF_ELEMENT_INDEX == "LootIcon" or DUF_ELEMENT_INDEX == "LeaderIcon" or DUF_ELEMENT_INDEX == "RoleIcon" or DUF_ELEMENT_INDEX == "PhaseIcon" or DUF_ELEMENT_INDEX == "HappinessIcon") then
		DUF_SPECIAL_OPTIONS = "DUF_IconOptions";
		DUF_Init_IconOptions();
	elseif (DUF_ELEMENT_INDEX == "Portrait") then
		DUF_SPECIAL_OPTIONS = "DUF_PortraitOptions";
		DUF_Init_PortraitOptions();
	elseif (DUF_ELEMENT_INDEX == "ClassIcon") then
		DUF_SPECIAL_OPTIONS = "DUF_ClassIconOptions";
		DUF_Init_ClassIconOptions();
	elseif (DUF_ELEMENT_INDEX == "TextBox") then
		DUF_SPECIAL_OPTIONS = "DUF_TextBoxOptions";
		DUF_Init_TextBoxOptions();
	elseif (DUF_ELEMENT_INDEX == "ROGUE") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Rogue_Settings";
		DUF_Init_Options_Rogue()
	elseif (DUF_ELEMENT_INDEX == "PALADIN") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Paladin_Settings"
		DUF_Init_Options_Paladin()
	elseif (DUF_ELEMENT_INDEX == "MONK") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Monk_Settings";
		DUF_Init_Options_Monk()
	elseif (DUF_ELEMENT_INDEX == "MONK"..SPEC_MONK_BREWMASTER) then -- Monk1 = Stagger
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Monk_Stagger_Settings";
		DUF_Init_Options_Monk_Stagger()
	elseif (DUF_ELEMENT_INDEX == "MAGE") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Mage_Settings";
		DUF_Init_Options_Mage()
	elseif (DUF_ELEMENT_INDEX == "DEATHKNIGHT") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_DeathKnight_Settings";
		DUF_Init_Options_DeathKnight()
	elseif (DUF_ELEMENT_INDEX == "DRUID") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Druid_Settings";
		DUF_Init_Options_Druid()
	elseif (DUF_ELEMENT_INDEX == "EVOKER") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Evoker_Settings";
		DUF_Init_Options_Evoker()
	elseif (DUF_ELEMENT_INDEX == "WARLOCK") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_Warlock_Settings";
		DUF_Init_Options_Warlock()
	elseif (DUF_ELEMENT_INDEX == "TotemFrame") then
		DUF_SPECIAL_OPTIONS = "DUF_ElementOptions_TotemFrame_Settings";
		DUF_Init_TotemFrame()
	elseif (DUF_ELEMENT_INDEX == "Buffs") then
		DUF_SPECIAL_OPTIONS = "DUF_BuffOptions";
		DUF_Init_BuffOptions();
	elseif (DUF_ELEMENT_INDEX == "Debuffs") then
		DUF_SPECIAL_OPTIONS = "DUF_DebuffOptions";
		DUF_Init_DebuffOptions();
	elseif (DUF_ELEMENT_INDEX == "EliteTexture") then
		DUF_SPECIAL_OPTIONS = "DUF_EliteTextureOptions";
		DUF_Init_EliteTextureOptions();
	elseif (not DUF_ELEMENT_INDEX) then
		DUF_SPECIAL_OPTIONS = "DUF_FrameOptions";
		DUF_FrameOptions:Show();
		DUF_Init_FrameOptions();
	elseif (DUF_ELEMENT_INDEX == "TempEnchants") then
		DUF_SPECIAL_OPTIONS = "DUF_TempEnchantOptions";
		DUF_Init_TempEnchantOptions();
	elseif (DUF_ELEMENT_INDEX == "Markers") then
		DUF_SPECIAL_OPTIONS = "DUF_TargetMarkersOptions";
		DUF_Init_TargetMarkersOptions();
	elseif (DUF_ELEMENT_INDEX == "QuestBossIcon") then
		DUF_SPECIAL_OPTIONS = "DUF_QuestBossIconOptions";
		DUF_Init_QuestBossIconOptions()
	end
	if (DUF_OPTIONS_SELECTED == "DUF_ElementOptions_SpecialSelect") then
		_G[DUF_SPECIAL_OPTIONS]:Show();
	end
end

local function DUF_SetMenu(index)
	if (index) then
		DUF_GENERALPROPERTIES = true;
		DUF_Options_FrameOptions:Hide()
		DUF_Options_Buffs:Hide()
		DUF_Options_ClassIcon:Hide()
		DUF_Options_Debuffs:Hide()
		DUF_Options_HealthBar:Hide()
		DUF_Options_LeaderIcon:Hide()
		DUF_Options_LootIcon:Hide()
		DUF_Options_ManaBar:Hide()
		DUF_Options_Portrait:Hide()
		DUF_Options_PVPIcon:Hide()
		DUF_Options_RaceIcon:Hide()
		DUF_Options_RankIcon:Hide()
		DUF_Options_RoleIcon:Hide()
		DUF_Options_StatusIcon:Hide()
		DUF_Options_PhaseIcon:Hide()
		DUF_Options_TargetHealthBar:Hide()
		DUF_Options_TargetManaBar:Hide()
		DUF_Options_CastBar:Hide()
		DUF_Options_TextBox1:Hide()
		DUF_Options_TextBox2:Hide()
		DUF_Options_TextBox3:Hide()
		DUF_Options_TextBox4:Hide()
		DUF_Options_TextBox5:Hide()
		DUF_Options_TextBox6:Hide()
		DUF_Options_TextBox7:Hide()
		DUF_Options_TextBox8:Hide()
		DUF_Options_TextBox9:Hide()
		DUF_Options_TextBox10:Hide()
		DUF_Options_XPBar:Hide()
		DUF_Options_HonorBar:Hide()
		DUF_Options_ClassProperties:Hide()

		DUF_Options_GeneralProperties:Show()
		DUF_Options_TempEnchantBar:Show()
		DUF_Options_AltManaBar:Show()
		DUF_Options_AltManaBarTxt:Show()
		DUF_Options_RogueCombo:Show();
		DUF_Options_DruidCombo:Show();
		DUF_Options_TotemFrame:Show()
		for i=1, #RemoveOptions.DUF_Options do
			local f = DUF_Options[RemoveOptions.DUF_Options[i].frame]
			if not f.VersionHidden then
				f:Show()
			end
		end
--[[			
		DUF_Options_Paladin:Show()
		DUF_Options_Monk:Show()
		DUF_Options_Monk_Stagger:Show()
		DUF_Options_Mage:Show();
		DUF_Options_DeathKnight:Show()
		DUF_Options_Warlock:Show()
]]--
	else
		DUF_GENERALPROPERTIES = nil;
		DUF_Options_FrameOptions:Show()
		DUF_Options_Buffs:Show()
		DUF_Options_ClassIcon:Show()
		DUF_Options_Debuffs:Show()
		DUF_Options_HealthBar:Show()
		DUF_Options_LeaderIcon:Show()
		DUF_Options_LootIcon:Show()
		DUF_Options_ManaBar:Show()
		DUF_Options_Portrait:Show()
		DUF_Options_PVPIcon:Show()
		DUF_Options_RaceIcon:Show()
		DUF_Options_RankIcon:Show()
		DUF_Options_RoleIcon:Show()
		DUF_Options_StatusIcon:Show()
		DUF_Options_PhaseIcon:Show()
		DUF_Options_TargetHealthBar:Show()
		DUF_Options_TargetManaBar:Show()
		DUF_Options_CastBar:Show()
		DUF_Options_TextBox1:Show()
		DUF_Options_TextBox2:Show()
		DUF_Options_TextBox3:Show()
		DUF_Options_TextBox4:Show()
		DUF_Options_TextBox5:Show()
		DUF_Options_TextBox6:Show()
		DUF_Options_TextBox7:Show()
		DUF_Options_TextBox8:Show()
		DUF_Options_TextBox9:Show()
		DUF_Options_TextBox10:Show()
		DUF_Options_XPBar:Show()
		DUF_Options_HonorBar:Show()
		DUF_Options_ClassProperties:Show()

		DUF_Options_GeneralProperties:Hide()
		DUF_Options_AltManaBar:Hide()
		DUF_Options_AltManaBarTxt:Hide()
		DUF_Options_TempEnchantBar:Hide()
		DUF_Options_RogueCombo:Hide();
		DUF_Options_Paladin:Hide()
		DUF_Options_Monk:Hide()
		DUF_Options_Monk_Stagger:Hide()
		DUF_Options_Mage:Hide();
		DUF_Options_DeathKnight:Hide()
		DUF_Options_DruidCombo:Hide();
		DUF_Options_Warlock:Hide()
		DUF_Options_Evoker:Hide()
		DUF_Options_TotemFrame:Hide()
	end
end

---------------------------------------------------------*** End of locals ***----------------------------------------------

function  DUF_Options_Initialize_StatusBar(self, ...)
	DUF_Initialize_StatusBar(...)
end

function DUF_Options_Initialize_TextBox(self, ...)
	DUF_Initialize_TextBox(...)
end

function DUF_Options_OnLoad(self)
	tinsert(UISpecialFrames, "DUF_Options");
	for parent, frames in pairs(RemoveOptions) do
		for i=1, #frames do
			if frames[i][WOW_PROJECT_ID] then
				local f = _G[parent][frames[i].frame]
				f:Hide()
				f.VersionHidden = true
			end
		end
	end
	for parent, frames in pairs(ReanchorOptions) do
		for i=1, #frames do
			if frames[i][WOW_PROJECT_ID] then
				local f = _G[parent][frames[i].frame]
				local tf = _G[parent][frames[i].toframe]
				local p, r, t, x, y = f:GetPoint(1)
				f:ClearAllPoints()
				f:SetPoint(p, tf, t, x, y)
			end
		end
	end
end

function DUF_Options_Initialize_TotemFrame(frame)
	DUF_Initialize_TotemFrame(true)
end

function DUF_Options_Initialize_ClassBar(frame)
	if DUF_ELEMENT_INDEX == "MONK"..SPEC_MONK_BREWMASTER then  -- Stagger
		if not ClassFramesSaved[DUF_ELEMENT_INDEX] then
			ClassFramesSaved[DUF_ELEMENT_INDEX] = {
				textures = {},
				borders = {},
			}
			ClassFramesSaved[DUF_ELEMENT_INDEX].textures[1] = DUF_Settings[DUF_INDEX].player[DUF_ELEMENT_INDEX].textures[1]
		end
		DUF_Initialize_ClassBar("MONK", SPEC_MONK_BREWMASTER)
		return
	end
	if not ClassFramesSaved[DUF_ELEMENT_INDEX] then
		if not DUF_Settings[DUF_INDEX].player[DUF_ELEMENT_INDEX].textures then return end
		frame.currentMax = #DUF_Settings[DUF_INDEX].player[DUF_ELEMENT_INDEX].textures
		ClassFramesSaved[DUF_ELEMENT_INDEX] = {
			textures = {},
			borders = {},
		}
		for i=1, DUF_CLASS_BAR_SPECS[DUF_ELEMENT_INDEX].total do
			if DUF_Settings[DUF_INDEX].player[DUF_ELEMENT_INDEX].textures[i] then
				ClassFramesSaved[DUF_ELEMENT_INDEX].textures[i] = DUF_Settings[DUF_INDEX].player[DUF_ELEMENT_INDEX].textures[i]
			else
				ClassFramesSaved[DUF_ELEMENT_INDEX].textures[i] = { texture="", }
			end
			if DUF_Settings[DUF_INDEX].player[DUF_ELEMENT_INDEX].borders[i] then
				ClassFramesSaved[DUF_ELEMENT_INDEX].borders[i] = DUF_Settings[DUF_INDEX].player[DUF_ELEMENT_INDEX].borders[i]
			else
				ClassFramesSaved[DUF_ELEMENT_INDEX].borders[i] = { texture="", }
			end
		end
	end
	DUF_Initialize_ClassBar(DUF_ELEMENT_INDEX, ElementSpecSettings)
end

function DUF_Init_MiscOptions()
	DL_Init_CheckBox(DUF_MiscOptions_HidePartyInRaid, DUF_Settings[DUF_INDEX].hidepartyinraid);
	DL_Init_CheckBox(DUF_MiscOptions_HideTargetTarget, DUF_Settings[DUF_INDEX].hidetargettarget);
	DL_Init_CheckBox(DUF_MiscOptions_HideBuffFrame, DUF_Settings[DUF_INDEX].hidebuffframe);
	DL_Init_CheckBox(DUF_MiscOptions_HideDebuffFrame, DUF_Settings[DUF_INDEX].hidedebuffframe);
	DL_Init_CheckBox(DUF_MiscOptions_PetBattleHide, DUF_Settings[DUF_INDEX].petbattlehide);
	DL_Init_CheckBox(DUF_MiscOptions_PetBattleHidePlayer, DUF_Settings[DUF_INDEX].petbattlehideplayer);
	DL_Init_CheckBox(DUF_MiscOptions_PetBattleHideParty, DUF_Settings[DUF_INDEX].petbattlehideparty);
	DL_Init_Slider(DUF_MiscOptions_UpdateSpeed, DUF_Settings[DUF_INDEX].updatespeedbase);
	DL_Init_CheckBox(DUF_MiscOptions_LimitDurationUpdates, DUF_Settings[DUF_INDEX].limitdurationupdates); 
	DL_Init_CheckBox(DUF_MiscOptions_CustomPaths, DUF_Settings[DUF_INDEX].custompaths); 
	DL_Init_EditBox(DUF_MiscOptions_RootPath, DUF_Settings[DUF_INDEX].rootpath);
	DL_Init_EditBox(DUF_MiscOptions_LowHealthThreshold, DUF_Settings[DUF_INDEX].flashthreshold);
	DL_Init_MenuControl(DUF_MiscOptions_OptionsScale, DUF_Settings[DUF_INDEX].optionscale);
end

function DUF_Init_ContextColors()
	for i = 1,6 do
		DL_Init_ColorPicker(_G["DUF_ColorContextOptions_Difficulty"..i], DUF_Settings[DUF_INDEX]["lvlcolor"..i]);
		DL_Init_ColorPicker(_G["DUF_ColorContextOptions_Reaction"..i], DUF_Settings[DUF_INDEX]["reactioncolor"..i]);
	end
	DL_Init_ColorPicker(DUF_ColorContextOptions_Health1, DUF_Settings[DUF_INDEX].healthcolor1);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Health2, DUF_Settings[DUF_INDEX].healthcolor2);
	DL_Init_EditBox(DUF_ColorContextOptions_HealthThreshold1, DUF_Settings[DUF_INDEX].healththreshold1);
	DL_Init_EditBox(DUF_ColorContextOptions_HealthThreshold2, DUF_Settings[DUF_INDEX].healththreshold2);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Mana1, DUF_Settings[DUF_INDEX].manacolor1);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Mana2, DUF_Settings[DUF_INDEX].manacolor2);
	DL_Init_EditBox(DUF_ColorContextOptions_ManaThreshold1, DUF_Settings[DUF_INDEX].manathreshold1);
	DL_Init_EditBox(DUF_ColorContextOptions_ManaThreshold2, DUF_Settings[DUF_INDEX].manathreshold2);
	DL_Init_ColorPicker(DUF_ColorContextOptions_DeathKnight, DUF_Settings[DUF_INDEX].classcolorDeathKnight);
	DL_Init_ColorPicker(DUF_ColorContextOptions_DemonHunter, DUF_Settings[DUF_INDEX].classcolorDemonHunter);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Druid, DUF_Settings[DUF_INDEX].classcolorDruid);
if EssencePlayerFrame then -- DragonFlight
	DL_Init_ColorPicker(DUF_ColorContextOptions_Evoker, DUF_Settings[DUF_INDEX].classcolorEvoker);
end
	DL_Init_ColorPicker(DUF_ColorContextOptions_Hunter, DUF_Settings[DUF_INDEX].classcolorHunter);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Mage, DUF_Settings[DUF_INDEX].classcolorMage);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Monk, DUF_Settings[DUF_INDEX].classcolorMonk);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Paladin, DUF_Settings[DUF_INDEX].classcolorPaladin);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Priest, DUF_Settings[DUF_INDEX].classcolorPriest);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Rogue, DUF_Settings[DUF_INDEX].classcolorRogue);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Shaman, DUF_Settings[DUF_INDEX].classcolorShaman);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Warlock, DUF_Settings[DUF_INDEX].classcolorWarlock);
	DL_Init_ColorPicker(DUF_ColorContextOptions_Warrior, DUF_Settings[DUF_INDEX].classcolorWarrior);
	DL_Init_CheckBox(DUF_ColorContextOptions_UseFor0, DUF_Settings[DUF_INDEX].usefor0);
	DL_Init_CheckBox(DUF_ColorContextOptions_UseFor1, DUF_Settings[DUF_INDEX].usefor1);
	DL_Init_CheckBox(DUF_ColorContextOptions_UseFor2, DUF_Settings[DUF_INDEX].usefor2);
	DL_Init_CheckBox(DUF_ColorContextOptions_UseFor3, DUF_Settings[DUF_INDEX].usefor3);
	DL_Init_CheckBox(DUF_ColorContextOptions_UseFor4, DUF_Settings[DUF_INDEX].usefor4);
end

local function SetClassColor(setting, class)
	local c = C_ClassColor.GetClassColor(class)
	setting.r, setting.g, setting.b, setting.a = c.r, c.g, c.b, c.a
end

function DUF_Set_DefaultClassColors()
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorDeathKnight, strupper("DeathKnight"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorDemonHunter, strupper("DemonHunter"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorDruid, strupper("Druid"))
if EssencePlayerFrame then -- DragonFlight
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorEvoker, strupper("Evoker"))
end
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorHunter, strupper("Hunter"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorMage, strupper("Mage"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorMonk, strupper("Monk"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorPaladin, strupper("Paladin"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorPriest, strupper("Priest"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorRogue, strupper("Rogue"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorShaman, strupper("Shaman"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorWarlock, strupper("Warlock"))
	SetClassColor(DUF_Settings[DUF_INDEX].classcolorWarrior, strupper("Warrior"))
	DUF_Init_ContextColors()
end

function DUF_CheckButton_OnClick(self)
 	DUF_Update_Setting(self:GetChecked(), self.index, self.subindex);
	if (self.index == "custompaths") then
	 	DUF_SetCustomPaths(self);
	end
	DUF_Run_Functions(self);
end

function DUF_SetCustomPaths()
	if (DUF_Settings[DUF_INDEX].custompaths) then
		DUF_MiscOptions_RootPath:Show();
		DUF_MiscOptions_SaveRoot:Show();
		DUF_MiscOptions_LoadRoots:Show();
	else
		DUF_MiscOptions_RootPath:Hide();
		DUF_MiscOptions_SaveRoot:Hide();
		DUF_MiscOptions_LoadRoots:Hide();
	end
end

function DUF_ColorPicker_OnClick(self)
	local basecolor = DUF_Options_GetSetting(self.index);
	if not basecolor then
		basecolor = { r=0, g=0, b=0 }
	end
	local color = {};
	color.r = basecolor.r or 0;
	color.g = basecolor.g or 0;
	color.b = basecolor.b or 0;
	ColorPickerFrame.previousValues = color;
	ColorPickerFrame.cancelFunc = DUF_ColorPicker_ColorCancelled;
	ColorPickerFrame.opacityFunc = DUF_ColorPicker_ColorChanged;
	ColorPickerFrame.func = DUF_ColorPicker_ColorChanged;
	ColorPickerFrame.initfunc = self:GetParent().initFunc;
	ColorPickerFrame.colorBox = self:GetName();
	ColorPickerFrame.index = self.index;
	ColorPickerFrame:SetColorRGB(color.r, color.g, color.b);
	DL_ColorBox_UpdateDisplay()
	ColorPickerFrame:ClearAllPoints();
	if (DUF_Options:GetRight() < UIParent:GetWidth() / 2) then
		ColorPickerFrame:SetPoint("LEFT", "DUF_Options", "RIGHT", 10, 0);
	else
		ColorPickerFrame:SetPoint("RIGHT", "DUF_Options", "LEFT", -10, 0);
	end
	ColorPickerFrame:Show();
end

function DUF_CopyFrame(self, source)
	if (not source) then return; end
	if (source == DUF_FRAME_INDEX) then return; end
--	local oldXPBar, oldCombo, oldElite, oldAniticpation; 
--	local oldHappiness
	local oldx = {};
	local oldy = {};
	DL_Copy_Table(DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].xos, oldx);
	DL_Copy_Table(DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].yos, oldy);

	local oldSettings = {}
	DL_Copy_Table(DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX], oldSettings);

	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX] = {};
	DL_Copy_Table(DUF_Settings[DUF_INDEX][source], DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX]);

	DL_Copy_Table(oldx, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].xos);
	DL_Copy_Table(oldy, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].yos);
--	if DUF_FRAME_INDEX == "pet" then
--		oldHappiness = {}
--		DL_Copy_Table(DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].HappinessIcon, oldHappiness)
--	end
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].EliteTexture = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].ROGUE = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].PALADIN = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].MONK = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX]["MONK"..SPEC_MONK_BREWMASTER] = nil -- Monk1 = Stagger
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].MAGE = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].DEATHKNIGHT = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].DRUID = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].WARLOCK = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].EVOKER = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].Markers = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].QuestBossIcon = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TempEnchants = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].StatusBar[3] = nil -- Player, Pet XPBar
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].StatusBar[7] = nil -- Fixed 5.3 -- Alt. Mana Bar
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].StatusBar[8] = nil -- Fixed 5.3 -- Honor Bar
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[11] = nil
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[12] = nil
	if (DUF_FRAME_INDEX == "player") then
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].ROGUE = {};
		DL_Copy_Table(oldSettings.ROGUE, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].ROGUE);
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].PALADIN = {}
		DL_Copy_Table(oldSettings.PALADIN, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].PALADIN)
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].MONK = {}
		DL_Copy_Table(oldSettings.MONK, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].MONK)
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX]["MONK"..SPEC_MONK_BREWMASTER] = {} -- Monk1 = Stagger
		DL_Copy_Table(oldSettings["MONK"..SPEC_MONK_BREWMASTER], DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX]["MONK"..SPEC_MONK_BREWMASTER]) -- Monk1 = Stagger
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].MAGE = {} 
		DL_Copy_Table(oldSettings.MAGE, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].MAGE)
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].DEATHKNIGHT = {}
		DL_Copy_Table(oldSettings.DEATHKNIGHT, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].DEATHKNIGHT)
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].EVOKER = {}
		DL_Copy_Table(oldSettings.EVOKER, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].EVOKER)
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].DRUID = {};
		DL_Copy_Table(oldSettings.DRUID, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].DRUID);
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].WARLOCK = {}
		DL_Copy_Table(oldSettings.WARLOCK, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].WARLOCK)
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].StatusBar[7] = {} --Druid Mana bar -- Fixed 5.3
		DL_Copy_Table(oldSettings.StatusBar[7], DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].StatusBar[7]) -- Fixed 5.3
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TempEnchants = {} 
		DL_Copy_Table(oldSettings.TempEnchants, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TempEnchants)
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[11] = {}
		DL_Copy_Table(oldSettings.TextBox[11], DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[11])
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[12] = {}
		DL_Copy_Table(oldSettings.TextBox[12], DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[12])
	end
	if (DUF_FRAME_INDEX ~= "pet") then
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].HappinessIcon = nil
	else
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].HappinessIcon = {}
		DL_Copy_Table(oldSettings.HappinessIcon, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].HappinessIcon)
	end
	if (DUF_FRAME_INDEX == "pet" or DUF_FRAME_INDEX == "player") then
		if (source ~= "player" and source ~= "pet") then
			DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].StatusBar[3] = {} -- XP bar
			DL_Copy_Table(oldSettings.StatusBar[3], DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].StatusBar[3])
		end
	end
	if (DUF_FRAME_INDEX == "target" or DUF_FRAME_INDEX == "boss") then
		if (source ~= "target" and source ~= "boss") then
			DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].EliteTexture = {};
			DL_Copy_Table(oldSettings.EliteTexture, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].EliteTexture);
		end
	end
	if (DUF_FRAME_INDEX == "target") then
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].Markers = {};
		DL_Copy_Table(oldSettings.Markers, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].Markers);
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].QuestBossIcon = {};
		DL_Copy_Table(oldSettings.QuestBossIcon, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].QuestBossIcon);
	end
	DUF_Initialize_AllFrames();
	DUF_Init_BaseOptions();
end

function DUF_EditBox_Update(self)
	local value = self:GetText();
	if (not value) then value = ""; end
	if (self.number) then
		value = self:GetNumber();
		if (not value) then
			value = 0;
			self:SetText("0");
		end
	end
	self:ClearFocus();
	if (self.index == "FRAMELOC") then
		DUF_EditBox_UpdateFrameLoc(self.subindex, value)
	else
		DUF_Update_Setting(value, self.index, self.subindex, self.unusualsubindex);
		DUF_Run_Functions(self);
		if (DUF_FRAME_INDEX) then
			for i = 1,10 do
				if (DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[i].label) then
					_G["DUF_Options_TextBox"..i]:SetText("["..i.."] "..DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[i].label);
				else
					_G["DUF_Options_TextBox"..i]:SetText(DUF_TEXT["TextBox"..i]);
				end
			end
		end
	end
end

function DUF_Options_Init_UnitFrameLoc(self, unit)
	local unitFrame = _G[DUF_FRAME_DATA[unit].frame]
	local settings = DUF_Settings[DUF_INDEX][DUF_FRAME_DATA[unit].index]
	unitFrame:SetPoint(settings.attachpoint[unitFrame:GetID()], UIParent, settings.attachto[unitFrame:GetID()], settings.xos[unitFrame:GetID()], settings.yos[unitFrame:GetID()])
end

function DUF_Init_Element(self, unit)
	if ElementClassSettings and ElementSpecSettings then
		DUF_Initialize_ClassBar(ElementClassSettings, ElementSpecSettings)
	elseif DUF_CLASS_BAR_SPECS[DUF_ELEMENT_INDEX] then
		DUF_Initialize_ClassBar(DUF_ELEMENT_INDEX)
	elseif (DUF_ELEMENT_INDEX == "Buffs") then
		DUF_Initialize_Buffs(self, unit);
	elseif (DUF_ELEMENT_INDEX == "ClassIcon") then
		DUF_Initialize_ClassIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "Debuffs") then
		DUF_Initialize_Debuffs(self, unit);
	elseif (DUF_ELEMENT_INDEX == "HappinessIcon") then
		DUF_Initialize_HappinessIcon();
	elseif (DUF_ELEMENT_INDEX == "LeaderIcon") then
		DUF_Initialize_LeaderIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "LootIcon") then
		DUF_Initialize_LootIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "Portrait") then
		DUF_Initialize_Portrait(unit);
	elseif (DUF_ELEMENT_INDEX == "PVPIcon") then
		DUF_Initialize_PVPIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "RaceIcon") then
		DUF_Initialize_RaceIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "RankIcon") then
		DUF_Initialize_RankIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "RoleIcon") then
		DUF_Initialize_RoleIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "StatusBar") then
		DUF_Initialize_StatusBar(unit, DUF_ELEMENT_SUBINDEX); 
	elseif (DUF_ELEMENT_INDEX == "StatusIcon") then
		DUF_Initialize_StatusIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "PhaseIcon") then
		DUF_Initialize_PhaseIcon(unit);
	elseif (DUF_ELEMENT_INDEX == "TextBox") then
		DUF_Initialize_TextBox(unit, DUF_ELEMENT_SUBINDEX); 
	elseif (DUF_ELEMENT_INDEX == "EliteTexture") then
		DUF_Initialize_EliteTexture();
	elseif (DUF_ELEMENT_INDEX == "TempEnchants") then
		DUF_Initialize_TempEnchants();
	elseif (DUF_ELEMENT_INDEX == "Markers") then
		DUF_Initialize_TargetMarkers(); 
	elseif (DUF_ELEMENT_INDEX == "QuestBossIcon") then
		DUF_Initialize_QuestBossIcon(); 
	elseif (DUF_ELEMENT_INDEX == "TotemFrame") then
		DUF_Initialize_TotemFrame(true);
	end
end

function DUF_Menu_OnClick(self)
	_G[DUF_DropMenu.controlbox]:SetText(_G[self:GetName().."_Text"]:GetText());
	DUF_DropMenu:Hide();
	if (DUF_DropMenu.index == "attach") then
		if (self.value and DUF_FRAME_DATA[DUF_FRAME_INDEX] and _G[DUF_FRAME_DATA[DUF_FRAME_INDEX].frame..self.value]) then
			local otherindex = _G[DUF_FRAME_DATA[DUF_FRAME_INDEX].frame..self.value].index;
			local othersubindex = _G[DUF_FRAME_DATA[DUF_FRAME_INDEX].frame..self.value].subindex;
			local otherattach;
			if (othersubindex) then
				otherattach = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][otherindex][othersubindex].attach;
			else
				otherattach = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][otherindex].attach;
			end
			otherattach = DUF_FRAME_DATA[DUF_FRAME_INDEX].frame..otherattach;
			local elemext = "_"..DUF_ELEMENT_INDEX;
			if (DUF_ELEMENT_SUBINDEX) then
				if (DUF_ELEMENT_INDEX == "StatusBar") then
					elemext = DUF_STATUSBARS[DUF_ELEMENT_SUBINDEX];
				elseif (DUF_ELEMENT_INDEX == "TextBox") then
					elemext = "_TextBox_"..DUF_ELEMENT_SUBINDEX;
				end
			end
			local element = DUF_FRAME_DATA[DUF_FRAME_INDEX].frame..elemext;
			if (element == otherattach ) then
				-- DL_Debug("That element is already attached to self element.  Cross-attaching the elements will crash WoW.  Attach option not applied.");
				return;
			end
		end
	end
	if DUF_DropMenu.index == "unitattach" then
		local frameindex = tonumber(DUF_FrameOptions_FrameChooser_Setting:GetText())
		DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX][DUF_DropMenu.subindex][frameindex] = self.value
	elseif (DUF_DropMenu.index) then
		DUF_Update_Setting(self.value, DUF_DropMenu.index);
	end
	DUF_Run_Functions(self, DUF_DropMenu.initFunc);
end

function DUF_Element_Nudge(self, button)
	if (button == "RightButton") then
		self.data[self.index] = self.data[self.index] + self.bumpval
	elseif (self.moving) then
		self.data[self.index] = self.data[self.index] + self.glideval
	else
		self.data[self.index] = self.data[self.index] + self.nudgeval
	end
	self.initfunc()
	DUF_Init_TargetMarkersOptions()
end

function DUF_Element_Nudge_OnUpdate(self, elapsed)
	if not self.timer then
		self.timer = .05
	end
	self.timer = self.timer - elapsed
	if self.timer < 0 then
		self.timer = .05
		if self.moving then
			DUF_Element_Nudge(self, button)
		end
	end
end

function DUF_Element_Nudge_Onload(self, label, data, index, nudgeval, bumpval, glideval)
	self:RegisterForClicks("LeftButtonUp", "RightButtonUp", "MiddleButtonUp");
	self:SetText(label)
	self.data = data
	self.index = index
	self.nudgeval = nudgeval
	self.bumpval = bumpval
	self.glideval = glideval
end

function DUF_Nudge_Down(self, button)
	local yoffset = DUF_Options_GetSetting("yoffset");
	if (button == "RightButton") then
		yoffset = yoffset - 10;
	elseif (self.moving) then
		yoffset = yoffset - 3;
	else
		yoffset = yoffset - 1;
	end
	DUF_ElementOptions_Base_YOffset:SetText(yoffset);
	DUF_Update_Setting(yoffset, "yoffset");
	DUF_Run_Functions(self);
end

function DUF_Nudge_Left(self, button)
	local xoffset = DUF_Options_GetSetting("xoffset");
	if (button == "RightButton") then
		xoffset = xoffset - 10;
	elseif (self.moving) then
		xoffset = xoffset - 3;
	else
		xoffset = xoffset - 1;
	end
	DUF_ElementOptions_Base_XOffset:SetText(xoffset);
	DUF_Update_Setting(xoffset, "xoffset");
	DUF_Run_Functions(self);
end

function DUF_Nudge_OnUpdate(self, elapsed)
	if (not self.timer) then
		self.timer = .05;
	end
	self.timer = self.timer - elapsed;
	if (self.timer < 0) then
		self.timer = .05;
		if (self.moving) then
			self.func(self);
		end
	end
end

function DUF_Nudge_Right(self, button)
	local xoffset = DUF_Options_GetSetting("xoffset");
	if (button == "RightButton") then
		xoffset = xoffset + 10;
	elseif (self.moving) then
		xoffset = xoffset + 3;
	else
		xoffset = xoffset + 1;
	end
	DUF_ElementOptions_Base_XOffset:SetText(xoffset);
	DUF_Update_Setting(xoffset, "xoffset");
	DUF_Run_Functions(self);
end

function DUF_Nudge_Up(self, button)
	local yoffset = DUF_Options_GetSetting("yoffset");
	if (button == "RightButton") then
		yoffset = yoffset + 10;
	elseif (self.moving) then
		yoffset = yoffset + 3;
	else
		yoffset = yoffset + 1;
	end
	DUF_ElementOptions_Base_YOffset:SetText(yoffset);
	DUF_Update_Setting(yoffset, "yoffset");
	DUF_Run_Functions(self);
end

function DUF_UnitFrame_Nudge_Up(self, button)
	local index = tonumber(DUF_FrameOptions_FrameChooser_Setting:GetText())
	local yoffset = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].yos[index]
	if (button == "RightButton") then
		yoffset = yoffset + 10;
	elseif (self.moving) then
		yoffset = yoffset + 3;
	else
		yoffset = yoffset + 1;
	end
	DUF_FrameOptions_YOffset:SetText(yoffset);
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].yos[index] = yoffset
	if (strfind(DUF_FRAME_INDEX, "party") or strfind(DUF_FRAME_INDEX, "boss")) then
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX..index)
	else
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX)
	end
end

function DUF_UnitFrame_Nudge_Down(self, button)
	local index = tonumber(DUF_FrameOptions_FrameChooser_Setting:GetText())
	local yoffset = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].yos[index]
	if (button == "RightButton") then
		yoffset = yoffset - 10;
	elseif (self.moving) then
		yoffset = yoffset - 3;
	else
		yoffset = yoffset - 1;
	end
	DUF_FrameOptions_YOffset:SetText(yoffset);
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].yos[index] = yoffset
	if (strfind(DUF_FRAME_INDEX, "party") or strfind(DUF_FRAME_INDEX, "boss")) then
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX..index)
	else
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX)
	end
end

function DUF_UnitFrame_Nudge_Left(self, button)
	local index = tonumber(DUF_FrameOptions_FrameChooser_Setting:GetText())
	local xoffset = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].xos[index]
	if (button == "RightButton") then
		xoffset = xoffset - 10;
	elseif (self.moving) then
		xoffset = xoffset - 3;
	else
		xoffset = xoffset - 1;
	end
	DUF_FrameOptions_XOffset:SetText(xoffset);
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].xos[index] = xoffset
	if (strfind(DUF_FRAME_INDEX, "party") or strfind(DUF_FRAME_INDEX, "boss")) then
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX..index)
	else
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX)
	end
end

function DUF_UnitFrame_Nudge_Right(self, button)
	local index = tonumber(DUF_FrameOptions_FrameChooser_Setting:GetText())
	local xoffset = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].xos[index]
	if (button == "RightButton") then
		xoffset = xoffset + 10;
	elseif (self.moving) then
		xoffset = xoffset + 3;
	else
		xoffset = xoffset + 1;
	end
	DUF_FrameOptions_XOffset:SetText(xoffset);
	DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].xos[index] = xoffset
	if (strfind(DUF_FRAME_INDEX, "party") or strfind(DUF_FRAME_INDEX, "boss")) then
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX..index)
	else
		DUF_Initialize_FrameLoc(DUF_FRAME_INDEX)
	end
end

function DUF_Options_UpdateFrameLoc()
	local index = tonumber(DUF_FrameOptions_FrameChooser_Setting:GetText())
	local yoffset = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].yos[index]
	local xoffset = DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].xos[index]
	DUF_FrameOptions_YOffset:SetText(yoffset)
	DUF_FrameOptions_XOffset:SetText(xoffset)
	DL_Init_MenuControl(DUF_FrameOptions_AttachPoint, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].attachpoint[index]);
	DL_Init_MenuControl(DUF_FrameOptions_AttachTo, DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].attachto[index]);
	return(index)
end

function DUF_Options_DeleteSettings(index)
	if (not index) then
		index = DUF_LOADINDEX;
	end
	if (index == "" or (not index)) then
		return;
	end
	if (index == DUF_INDEX) then
		DL_Error("You cannot delete the profile you're currently using.");
		return;
	end
	if (not DUF_Settings[index]) then
		DL_Feedback(DUF_TEXT.NoSettingsFound);
		return;
	end
	DUF_Settings[index] = nil;
	DUF_Update_SavedSettings();
	DUF_MiscOptions_LoadSettings_Setting:SetText("");
	DL_Feedback(DUF_TEXT.SettingsDeleted);
end

function DUF_Options_FrameLevelMinus(self)
	local level = DUF_Options_GetSetting("framelevel");
	level = level - 1;
	DUF_Update_Setting(level, "framelevel");
	DL_Init_EditBox(DUF_ElementOptions_Base_FrameLevel, level);
	DUF_Run_Functions(self, DUF_ElementOptions_Base_FrameLevel.initFunc);
end

function DUF_Options_FrameLevelPlus(self)
	local level = DUF_Options_GetSetting("framelevel");
	level = level + 1;
	DUF_Update_Setting(level, "framelevel");
	DL_Init_EditBox(DUF_ElementOptions_Base_FrameLevel, level);
	DUF_Run_Functions(self, DUF_ElementOptions_Base_FrameLevel.initFunc);
end

function DUF_Options_OnHide(self)
	DUF_OPTIONS_VISIBLE = nil;
	for unit, value in pairs(DUF_FRAME_DATA) do
		if not DUF_FRAME_NOP_VERSIONS[value.index] then
			if (not DUF_Settings[DUF_INDEX][value.index].hide) then
     					local frame = _G[value.frame];
				DUF_SetFrameUnitWatch(frame, true)
			end
		end
	end
	DUF_Main_UpdatePartyMembers();

	if (DUF_Settings[DUF_INDEX].autolock) then
		if (DUF_FRAMES_UNLOCKED) then
			DUF_Toggle_FramesLock();
		end
		if (DUF_ELEMENTS_UNLOCKED) then
			DUF_Toggle_ElementsLock();
		end
	end
	DUF_Initialize_AllFrames();
end

function DUF_Options_OnShow(self)
	if (not DUF_INITIALIZED) then return; end
	DUF_OPTIONS_VISIBLE = true;
	DUF_Initialize_AllFrames();
	DUF_Options_ShowFrames()
	if not self.Init then
		self.Init = true
		DUF_Options_Self:Click()
	end
end

function DUF_Options_ShowFrames()
	for unit, value in pairs(DUF_FRAME_DATA) do
		if not DUF_FRAME_NOP_VERSIONS[value.index] then
			if (not DUF_Settings[DUF_INDEX][value.index].hide) then
				local frame = _G[value.frame]
				DUF_SetFrameUnitWatch(frame, false)
				frame:Show()
			end
		end
	end
end

function DUF_Options_OnUpdate(self)
	if (not DUF_INITIALIZED) then return; end
	if (not self.scale) then
		self.scale = self:GetScale();
	end
	if (self.scale ~= self:GetScale()) then
		DUF_Options:SetScale(DUF_Settings[DUF_INDEX].optionscale);
		self.scale = self:GetScale();
	end
end

function DUF_Options_SaveSettings(index)
	if (not index) then
		index = DUF_MiscOptions_SaveSettings:GetText();
	end
	DUF_MiscOptions_SaveSettings:ClearFocus();
	DUF_MiscOptions_SaveSettings:SetText("");
	if (index == "" or (not index)) then
		return;
	end
	if (DUF_Settings[index]) then
		DL_Error("You are attempting to create a new profile using the same name as an existing profile.  You must delete the existing profile first.");
		return;
	end
	DUF_Settings[index] = {};
	DL_Copy_Table(DUF_Settings[DUF_INDEX], DUF_Settings[index]);
	DUF_INDEX = index;
	DUF_Settings[DUF_PLAYER_INDEX] = index;
	DUF_Update_SavedSettings();
	DL_Feedback(DUF_TEXT.ProfileCreated);
end

function DUF_Options_SelectFrameOptions()
	DUF_Options_ResetIndices();
	DUF_MiscOptions:Hide();
	DUF_ColorContextOptions:Hide();
	DUF_ElementOptions:Hide();
	if (DUF_ELEMENT_SELECTED) then
		local button = _G[DUF_ELEMENT_SELECTED];
--		button:SetNormalFontObject(GameFontNormal);		
--		button:GetNormalFontObject():SetTextColor(1, 1, 0);
		button.Text:SetFontObject(GameFontNormal);		
		button.Text:SetTextColor(1, 1, 0);
		button:SetBackdropColor(0, 0, 0);
		button:SetBackdropBorderColor(.8, 0, 0);
	end
	DUF_ELEMENT_INDEX = nil;
	DUF_ELEMENT_SUBINDEX = nil;
	DUF_ELEMENT_SELECTED = "DUF_Options_FrameOptions";
--	DUF_Options_FrameOptions:SetNormalFontObject(GameFontNormal);
--	DUF_Options_FrameOptions:GetNormalFontObject():SetTextColor(1, 0, 0);
	DUF_Options_FrameOptions.Text:SetFontObject(GameFontNormal);
	DUF_Options_FrameOptions.Text:SetTextColor(1, 0, 0);
	DUF_Options_FrameOptions:SetBackdropColor(1, 1, 0);
	DUF_Options_FrameOptions:SetBackdropBorderColor(1, 1, 1);
end

function DUF_Options_SetLoadIndex()
	DUF_LOADINDEX = DUF_MiscOptions_LoadSettings_Setting:GetText();
end

function DUF_Options_SetRootIndex()
	DUF_Settings[DUF_INDEX].rootpath = DUF_MiscOptions_LoadRoots_Setting:GetText();
	DUF_MiscOptions_RootPath:SetText(DUF_MiscOptions_LoadRoots_Setting:GetText());
	DUF_Options_LoadSettings(DUF_INDEX);
end

function DUF_Options_SetAttachPoint()
	DUF_Settings[DUF_INDEX].attachpoint = DUF_MiscOptions_LoadRoots_Setting:GetText();
	DUF_MiscOptions_RootPath:SetText(DUF_MiscOptions_LoadRoots_Setting:GetText());
	DUF_Options_LoadSettings(DUF_INDEX);
end

function DUF_Options_SetScale()
	DUF_Options:SetScale(DUF_Settings[DUF_INDEX].optionscale);
	DUF_Options.scale = DUF_Options:GetScale();
	DUF_Options:ClearAllPoints();
	DUF_Options:SetPoint("CENTER", "UIParent", "CENTER", 0, 0);
end

function DUF_OptionsSelect_OnClick(self)
	if (DUF_OPTIONS_SELECTED) then
		local button = _G[DUF_OPTIONS_SELECTED];
--		button:SetNormalFontObject(GameFontNormal);		
--		button:GetNormalFontObject():SetTextColor(1, 1, 0);
		button.Text:SetFontObject(GameFontNormal);		
		button.Text:SetTextColor(1, 1, 0);

		button:SetBackdropColor(0, 0, 0);
		button:SetBackdropBorderColor(.8, 0, 0);
	end
	DUF_OPTIONS_SELECTED = self:GetName();
--	self:SetNormalFontObject(GameFontNormal);	
--	self:GetNormalFontObject():SetTextColor(1, 0, 0);
	self.Text:SetFontObject(GameFontNormal);	
	self.Text:SetTextColor(1, 0, 0);
	self:SetBackdropColor(1, 1, 0);
	self:SetBackdropBorderColor(1, 1, 1);
end

function DUF_SelectButton_OnLoad(self, passive)
--	self:SetNormalFontObject(GameFontNormal);	
--	self:GetNormalFontObject():SetTextColor(1, 1, 0);
	self.Text:SetFontObject(GameFontNormal);	
	self.Text:SetTextColor(1, 1, 0);
	self:SetBackdropColor(0, 0, 0);
	self:SetBackdropBorderColor(.8, 0, 0);
end

function DUF_SelectElement_OnClick(self, index, subindex)
	DUF_Options_ResetIndices();
	DUF_MiscOptions:Hide();
	DUF_ColorContextOptions:Hide();
	if (DUF_ELEMENT_SELECTED) then
		local button = _G[DUF_ELEMENT_SELECTED];
--		button:SetNormalFontObject(GameFontNormal);		
--		button:GetNormalFontObject():SetTextColor(1, 1, 0);
		button.Text:SetFontObject(GameFontNormal);		
		button.Text:SetTextColor(1, 1, 0);
		button:SetBackdropColor(0, 0, 0);
		button:SetBackdropBorderColor(.8, 0, 0);
	end
	DUF_ELEMENT_SELECTED = self:GetName();
--	self:SetNormalFontObject(GameFontNormal);	
--	self:GetNormalFontObject():SetTextColor(1, 0, 0);
	self.Text:SetFontObject(GameFontNormal);	
	self.Text:SetTextColor(1, 0, 0);
	self:SetBackdropColor(1, 1, 0);
	self:SetBackdropBorderColor(1, 1, 1);
	DUF_ELEMENT_INDEX = index;
	DUF_ELEMENT_SUBINDEX = subindex;
	if (DUF_ELEMENT_INDEX) then
		DUF_ElementOptions:Show();
		DUF_FrameOptions:Hide();
		DUF_Init_BaseOptions();
		DUF_Show_SpecialOptions();
	else
		DUF_ElementOptions:Hide();
		DUF_FrameOptions:Show();
		DUF_Init_FrameOptions();
	end
end

function DUF_SelectFrame_OnClick(self, index, bttn)
	if (not bttn) then
		bttn = self
	end
	DUF_Options_ResetIndices();
	DUF_MiscOptions:Hide();
	DUF_ColorContextOptions:Hide();
	if (DUF_FRAME_SELECTED) then
		local button = _G[DUF_FRAME_SELECTED];
--		button:SetNormalFontObject(GameFontNormal);
--		button:GetNormalFontObject():SetTextColor(1, 1, 0);
		button.Text:SetFontObject(GameFontNormal);
		button.Text:SetTextColor(1, 1, 0);
		button:SetBackdropColor(0, 0, 0);
		button:SetBackdropBorderColor(.8, 0, 0);
	end
	DUF_FRAME_SELECTED = bttn:GetName();
--	bttn:SetNormalFontObject(GameFontNormal);	
--	bttn:GetNormalFontObject():SetTextColor(1, 0, 0);
	bttn.Text:SetFontObject(GameFontNormal);	
	bttn.Text:SetTextColor(1, 0, 0);
	bttn:SetBackdropColor(1, 1, 0);
	bttn:SetBackdropColor(1, 1, 0);
	bttn:SetBackdropBorderColor(1, 1, 1);
	DUF_FRAME_INDEX = index;
	table.wipe(DUF_INIT_FRAMES)
	tinsert(DUF_INIT_FRAMES, index)
	DUF_SetMenu(nil)
	DUF_Options_XPBar:Hide();
	DUF_Options_HonorBar:Hide();
	DUF_Options_HappinessIcon:Hide();
	DUF_Options_EliteTexture:Hide();
	DUF_Options_TargetMarkers:Hide();
	DUF_Options_QuestBossIcon:Hide();
	DUF_MiscOptions:Hide();
	DUF_Options_ClassProperties:Hide();
	DUF_Options_GeneralProperties:Hide();
	DUF_ColorContextOptions:Hide();
	local Count = 4
	if index == "party" or index == "partypet" or index == "boss" then
		if index == "boss" then
			Count = 5
			DUF_Options_EliteTexture:Show()
		end
		DUF_INIT_FRAMES[1] = index.."1"
		for i=2, Count do
			tinsert(DUF_INIT_FRAMES, index..i)
		end
--		DUF_INIT_FRAMES = { "party1", "party2", "party3", "party4" };
--	elseif (index == "partypet") then
--		DUF_INIT_FRAMES = { "partypet1", "partypet2", "partypet3", "partypet4" };
--	elseif index == "boss" then
--		DUF_INIT_FRAMES[1] = index.."1"
--		for i=2, 5 do
--			tinsert(DUF_INIT_FRAMES, [index..i])
--		end
--		DUF_INIT_FRAMES = { "boss1", "boss2", "boss3", "boss4", "boss5" };
--		DUF_Options_EliteTexture:Show()
	elseif index == "player" then
		DUF_Options_XPBar:Show()
		DUF_Options_HonorBar:Show()
		DUF_Options_ClassProperties:Show()
	elseif index == "pet" then
		DUF_Options_XPBar:Show()
		DUF_Options_HappinessIcon:Show()
	elseif index == "target" then
		DUF_Options_EliteTexture:Show()
		DUF_Options_TargetMarkers:Show()
		DUF_Options_QuestBossIcon:Show()
	end

	table.wipe(DUF_FRAME_CHOICES)
	tinsert(DUF_FRAME_CHOICES, { value = 1, text = 1 })
	if index == "party" or index == "partypet" or index == "boss" then
		tinsert(DUF_FRAME_CHOICES, { value = 2, text = 2 })
		tinsert(DUF_FRAME_CHOICES, { value = 3, text = 3 })
		tinsert(DUF_FRAME_CHOICES, { value = 4, text = 4 })
	end
	if index == "boss" then
		tinsert(DUF_FRAME_CHOICES, { value = 5, text = 5 })
	end
	if (DUF_ELEMENT_INDEX == "StatusBar" and DUF_ELEMENT_SUBINDEX == 3 and DUF_FRAME_INDEX ~= "player" and DUF_FRAME_INDEX ~= "pet") then
		DUF_Options_SelectFrameOptions();
	elseif (DUF_ELEMENT_INDEX == "StatusBar" and (DUF_ELEMENT_SUBINDEX == 7) and DUF_FRAME_INDEX ~= "player") then -- Druid Mana is DUF_ELEMENT_SUBINDEX 7 -- Fixed 5.3
		DUF_Options_SelectFrameOptions();
	elseif ((DUF_ELEMENT_INDEX == "TempEnchants") and DUF_FRAME_INDEX ~= "player") then
		DUF_Options_SelectFrameOptions();
	elseif (DUF_ELEMENT_INDEX == "HappinessIcon" and DUF_FRAME_INDEX ~= "pet") then
		DUF_Options_SelectFrameOptions();
	elseif DUF_CLASS_BAR_SPECS[DUF_ELEMENT_INDEX] and DUF_FRAME_INDEX ~= "player" then
		if WOW_PROJECT_ID == WOW_PROJECT_MAINLINE then
			DUF_Options_SelectFrameOptions()
		end
	elseif ((DUF_ELEMENT_INDEX == "TextBox" and DUF_ELEMENT_SUBINDEX > 10) and DUF_FRAME_INDEX ~= "player") then
		DUF_Options_SelectFrameOptions();
	elseif (DUF_ELEMENT_INDEX == "EliteTexture" and DUF_FRAME_INDEX ~= "target") then
		DUF_Options_SelectFrameOptions();
	elseif (DUF_ELEMENT_INDEX == "Markers" and DUF_FRAME_INDEX ~= "target") then
		DUF_Options_SelectFrameOptions();
	elseif (DUF_ELEMENT_INDEX == "QuestBossIcon" and DUF_FRAME_INDEX ~= "target") then
		DUF_Options_SelectFrameOptions();
	end
	DUF_Show_SpecialOptions();
	if (DUF_ELEMENT_INDEX) then
		DUF_Init_BaseOptions();
	else
		DUF_Init_FrameOptions();
	end
	for i = 1,10 do
		_G["DUF_Options_TextBox"..i]:SetText("["..i.."] "..DUF_Settings[DUF_INDEX][DUF_FRAME_INDEX].TextBox[i].label);
	end
end

function DUF_SetCopyFrame(self)
	DUF_FrameOptions_CopyButton.index = self.value;
	DUF_FrameOptions_SafeCopyButton.index = self.value;
end

function DUF_Show_ContextColors()
	DUF_MiscOptions:Hide();
	DUF_ColorContextOptions:Show();
	DUF_ElementOptions:Hide();
	DUF_FrameOptions:Hide();
	DUF_Options_SetOldIndices();
end

function DUF_Show_MiscOptions()
	DUF_MiscOptions:Show();
	DUF_ColorContextOptions:Hide();
	DUF_ElementOptions:Hide();
	DUF_FrameOptions:Hide();
	DUF_Options_SetOldIndices();
end

function DUF_Slider_SyncLock(self)
	if not self.synclock then return end
	if not self.synclock:GetChecked() then return end
	local value = self:GetValue()
	local othervalue = self.syncframe:GetValue()
	if value > 0 and othervalue > 0 and value ~= othervalue then
		self.syncframe:SetValue(value)
		DUF_Slider_Update(self.syncframe)
	end
end

function DUF_Slider_Update(self)
	if (not DUF_INITIALIZED) then return; end
	local setting = DUF_Options_GetSetting(self.index);
	if (self.scale) then
		setting = setting * self.scale;
	end
	local min, max = self:GetMinMaxValues();
	if (setting and (setting < min or setting > max)) then
		return;
	end
	local value = self:GetValue();
	_G[self:GetName().."_Display"]:SetText(value);
	if (self.scale) then
		value = value / self.scale;
	end
	DUF_Update_Setting(value, self.index, self.subindex);
	DUF_Run_Functions(self);
	DUF_Slider_SyncLock(self)
end

function DUF_Slider_UpdateFromEditBox(self)
	if (DUF_NOSLIDEREDIT) then return; end
	local value = self:GetNumber();
	if (not value) then value = 0; end
	local min, max = self:GetParent():GetMinMaxValues();
	if (self:GetParent().minlock and value < min) then 
		value = min 
	end
	if (self:GetParent().maxlock and value > max) then 
		value = max 
	end
	self:SetText(value);
	if (value >= min and value <= max) then
		self:GetParent():SetValue(value);
	end
	self:ClearFocus();
	if (self:GetParent().scale) then
		value = value / self:GetParent().scale;
	end
	DUF_Update_Setting(value, self:GetParent().index, self:GetParent().subindex);
	if (self:GetParent().initFunc) then
		DUF_Run_Functions(self, self:GetParent().initFunc);
	else
		DUF_Run_Functions(self, self:GetParent():GetParent().initFunc);
	end
	DUF_Slider_SyncLock(self)
end

function DUF_Update_SavedSettings()
	DUF_SAVED_SETTINGS = {};
	for setting in pairs(DUF_Settings) do
		if ((not strfind(setting, "INITIALIZED")) and (not strfind(setting, " :: "))) then
			DUF_SAVED_SETTINGS[setting] = { text=setting, value=setting };
		end
	end
	DUF_MiscOptions_CurrentProfile:SetText(DUF_TEXT.CurrentProfile..":  "..DUF_INDEX);
end

function DUF_ShowProperties_OnClick(self)
	if (self:GetName() == "DUF_Options_GeneralProperties") then
		DUF_SetMenu(nil);
	else
		DUF_SetMenu(true);
	end
end

function DUF_Options_SaveRoot()
	if (DUF_Settings[DUF_INDEX].rootpath == "") then
		return;
	end
	local j = #DUF_Settings[DUF_INDEX].savedroots;
	if (j and j >0) then
		for i=1, j do
			if (strupper(DUF_Settings[DUF_INDEX].savedroots[i].value) == strupper(DUF_Settings[DUF_INDEX].rootpath)) then
				DL_Error(DUF_TEXT.RootSaveError);
				return;
			end
		end
	end
	j = j+1;
	DUF_Settings[DUF_INDEX]["savedroots"][j] = {};
	DUF_Settings[DUF_INDEX]["savedroots"][j]["text"] = DUF_Settings[DUF_INDEX].rootpath;
	DUF_Settings[DUF_INDEX]["savedroots"][j]["value"] = DUF_Settings[DUF_INDEX].rootpath;
end

-------------------------------- Text Variables List --------------------------------

local LookupTable = {} -- Only display in list variables we actually have functions for (some are version specific).
local Exempt_TEXTVARIABLES = { -- We may change these so don't "advertise" them atm.
	["$PL"] = true,
	["$NP"] = true,
}
for k, v in pairs(DUF_VARIABLE_FUNCTIONS) do
	if not Exempt_TEXTVARIABLES[k] then
		if not DUF_TEXTVARIABLES[k] then
			if FIZZLEDEBUG then
				print("DUF: Variables key missing from DUF_TEXTVARIABLES", k)
			end
		else	
			tinsert(LookupTable, { varsort=strlower(k), variable=k, text=DUF_TEXTVARIABLES[k], })
		end
	end
end

local SortKey
local function DoSort(a, b)
	return a[SortKey] < b[SortKey]
end
local function SortLookup(key)
	SortKey = key
	sort(LookupTable, DoSort)
end
SortLookup("varsort")
local numRows
local Orders = {
	[1] = { left="variable", right="text", colorL={ YELLOW_FONT_COLOR.r, YELLOW_FONT_COLOR.g, YELLOW_FONT_COLOR.b }, colorR={ 1, 1, 1, }, highlightL={ 0.0, 1, 0.0 }, highlightR={ 0.667, 1, 0.659, }, },
	[2] = { left="text", right="variable",  colorL={ 1, 1, 1, }, colorR={  YELLOW_FONT_COLOR.r, YELLOW_FONT_COLOR.g, YELLOW_FONT_COLOR.b }, highlightL={ 0.667, 1, 0.659, }, highlightR={ 0.0, 1, 0.0 }, },
}

function DUF_VarList_OnLoad(self)
	numRows = self:GetHeight() / 16
	self.Buttons = {}
	for  i=1, numRows do
		tinsert(self.Buttons, CreateFrame("Button", "$parentButton".. i, self, "DUF_TextVariableButtonTemplate"))
		local Button = self.Buttons[#self.Buttons]
		Button.Selected:SetStartPoint("BOTTOMLEFT", 10, 0)
		Button.Selected:SetEndPoint("BOTTOMRIGHT", -10, 0)
		Button.Selected:SetAlpha(0.5)
		Button:SetID(i)
		if i == 1 then
			Button:SetPoint("TOPLEFT", 2, 0)
			Button:SetPoint("TOPRIGHT", -2, 0)
		else
			Button:SetPoint("TOPLEFT", self.Buttons[#self.Buttons-1], "BOTTOMLEFT")
			Button:SetPoint("TOPRIGHT", self.Buttons[#self.Buttons-1], "BOTTOMRIGHT")
		end
	end
	local varwidth = 0
	local textwidth = 0
	local Button = self.Buttons[1]
	for k, v in pairs(LookupTable) do -- Determine max. width of $var text
		Button.Left:SetText(v.variable)
		local w = Button.Left:GetWidth()
		if w > varwidth then
			varwidth = w
		end
	end
	Orders[1].width = varwidth -- Set max. width of $var field
	Orders[2].width = Button:GetWidth() - (varwidth+6) -- Set max. width of text field
	self.Order = 1 
	DUF_VarList_Update(self)
end


function DUF_VarList_Update(self) --function DART_ControlMenu_Update()
	local list = LookupTable
	if (not list) then list = {} end
	self = self
	local numRows = #self.Buttons
	local height = self.Buttons[1]:GetHeight()
	local offset = FauxScrollFrame_GetOffset(self)
	if (not offset) then offset = 0 end
	local index, button, text
	for i=1, numRows do
		button = self.Buttons[i]
		index = offset + i
		if list[index] then
			button:Show()
			button.index = index
			local L = list[index][Orders[self.Order].left]
			local R = list[index][Orders[self.Order].right]
			local Col = "|cff707000$|r"
			if self.Order == 1 then
				L = Col..strsub(L, 2)
				button.Left:SetFontObject(GameFontNormalLarge)
				button.Right:SetFontObject(GameFontNormal)
			else
				button.Left:SetFontObject(GameFontNormal)
				button.Right:SetFontObject(GameFontNormalLarge)
				R = Col..strsub(R, 2)
			end
--			button.Left:SetText(list[index][Orders[self.Order].left])
--			button.Right:SetText(list[index][Orders[self.Order].right])
			button.Left:SetText(L)
			button.Right:SetText(R)
			button.Selected:Hide()
			if self.LastIndex and button.index == self.LastIndex then
				button.Left:SetTextColor(unpack(Orders[self.Order].highlightL))
				button.Right:SetTextColor(unpack(Orders[self.Order].highlightR))
				button.Selected:Show()
			else
				button.Left:SetTextColor(unpack(Orders[self.Order].colorL))
				button.Right:SetTextColor(unpack(Orders[self.Order].colorR))
			end
			button.Left:SetWidth(Orders[self.Order].width)
		else
			button:Hide();
		end
	end
	FauxScrollFrame_Update(self, #list, numRows, height);
end

function DUF_VarList_SetOrder(self)
	local Checked = self:GetChecked()
	local Parent = self:GetParent()
	Parent.Order = Checked and 2 or 1
	self:SetText(Checked and "$" or "...")
	local lastVar
	if Parent.LastIndex then
		lastVar = LookupTable[Parent.LastIndex].variable
	end
	SortLookup(Checked and "text" or "varsort")
	if Parent.LastIndex then
		for k, v in ipairs(LookupTable) do
			if v.variable == lastVar then
				Parent.LastIndex = k
				break
			end
		end
	end
	DUF_VarList_Update(Parent)
end

function DUF_VarList_InsertVariable_OnClick(self, button, down)
	DUF_TextBoxOptions_Text:SetFocus()
	local pos = DUF_TextBoxOptions_Text:GetCursorPosition()
	local t = DUF_TextBoxOptions_Text:GetText()
	if pos == 1 then
		t = self:GetText() .. t
	else
		t = strsub(t, 1, pos) .. self:GetText() .. strsub(t, pos + 1)
	end
	DUF_TextBoxOptions_Text:SetText(t)
	DUF_TextBoxOptions_Text:SetCursorPosition(pos + 3)
end


function DUF_VarList_OnClick(self, button, down)
	local Parent = self:GetParent()
	if Parent.LastIndex and Parent.LastIndex == self.index then 
		Parent.LastIndex = nil
		Parent.InsertVariable:Hide()
	else
		Parent.LastIndex = self.index
		Parent.InsertVariable:SetText(LookupTable[Parent.LastIndex].variable)
		Parent.InsertVariable:Show()
		DUF_TextBoxOptions_Text:SetFocus()
	end
	DUF_VarList_Update(Parent)
end
